import pandas as pd
import numpy as np
import re
from . import register_class, EXP_NAMESPACE
from . import get_type_map
from ..container import Table, Row, Container, AbstractContainer
from ..utils import docval, popargs, AllowPositional
from ..build import TypeMap


class KeyTable(Table):
    """
    A table for storing keys used to reference external resources.
    """

    __defaultname__ = 'keys'

    __columns__ = (
        {'name': 'key', 'type': str,
         'doc': 'The user key that maps to the resource term / registry symbol.'},
    )


class Key(Row):
    """
    A Row class for representing rows in the KeyTable.
    """

    __table__ = KeyTable


class ResourceTable(Table):
    """
    A table for storing names and URIs of ontology sources.
    """

    __defaultname__ = 'resources'

    __columns__ = (
        {'name': 'resource', 'type': str,
         'doc': 'The resource/registry that the term/symbol comes from.'},
        {'name': 'resource_uri', 'type': str,
         'doc': 'The URI for the resource term / registry symbol.'},
    )


class Resource(Row):
    """
    A Row class for representing rows in the ResourceTable.
    """

    __table__ = ResourceTable


class EntityTable(Table):
    """
    A table for storing the external resources a key refers to.
    """

    __defaultname__ = 'entities'

    __columns__ = (
        {'name': 'keys_idx', 'type': (int, Key),
         'doc': ('The index into the keys table for the user key that '
                 'maps to the resource term / registry symbol.')},
        {'name': 'resources_idx', 'type': (int, Resource),
         'doc': 'The index into the ResourceTable.'},
        {'name': 'entity_id', 'type': str,
         'doc': 'The unique ID for the resource term / registry symbol.'},
        {'name': 'entity_uri', 'type': str,
         'doc': 'The URI for the resource term / registry symbol.'},
    )


class Entity(Row):
    """
    A Row class for representing rows in the EntityTable.
    """

    __table__ = EntityTable


class ObjectTable(Table):
    """
    A table for storing objects (i.e. Containers) that contain keys that refer to external resources.
    """

    __defaultname__ = 'objects'

    __columns__ = (
        {'name': 'object_id', 'type': str,
         'doc': 'The object ID for the Container/Data.'},
        {'name': 'relative_path', 'type': str,
         'doc': ('The relative_path of the attribute of the object that uses ',
                 'an external resource reference key. Use an empty string if not applicable.')},
        {'name': 'field', 'type': str,
         'doc': ('The field of the compound data type using an external resource. '
                 'Use an empty string if not applicable.')}
    )


class Object(Row):
    """
    A Row class for representing rows in the ObjectTable.
    """

    __table__ = ObjectTable


class ObjectKeyTable(Table):
    """
    A table for identifying which keys are used by which objects for referring to external resources.
    """

    __defaultname__ = 'object_keys'

    __columns__ = (
        {'name': 'objects_idx', 'type': (int, Object),
         'doc': 'The index into the objects table for the Object that uses the Key.'},
        {'name': 'keys_idx', 'type': (int, Key),
         'doc': 'The index into the keys table that is used to make an external resource reference.'}
    )


class ObjectKey(Row):
    """
    A Row class for representing rows in the ObjectKeyTable.
    """

    __table__ = ObjectKeyTable


@register_class('ExternalResources', EXP_NAMESPACE)
class ExternalResources(Container):
    """A table for mapping user terms (i.e. keys) to resource entities."""

    __fields__ = (
        {'name': 'keys', 'child': True},
        {'name': 'resources', 'child': True},
        {'name': 'objects', 'child': True},
        {'name': 'object_keys', 'child': True},
        {'name': 'entities', 'child': True},
    )

    @docval({'name': 'name', 'type': str, 'doc': 'The name of this ExternalResources container.'},
            {'name': 'keys', 'type': KeyTable, 'default': None,
             'doc': 'The table storing user keys for referencing resources.'},
            {'name': 'resources', 'type': ResourceTable, 'default': None,
             'doc': 'The table for storing names and URIs of resources.'},
            {'name': 'entities', 'type': EntityTable, 'default': None,
             'doc': 'The table storing entity information.'},
            {'name': 'objects', 'type': ObjectTable, 'default': None,
             'doc': 'The table storing object information.'},
            {'name': 'object_keys', 'type': ObjectKeyTable, 'default': None,
             'doc': 'The table storing object-resource relationships.'},
            {'name': 'type_map', 'type': TypeMap, 'default': None,
             'doc': 'The type map. If None is provided, the HDMF-common type map will be used.'},
            allow_positional=AllowPositional.WARNING)
    def __init__(self, **kwargs):
        name = popargs('name', kwargs)
        super().__init__(name)
        self.keys = kwargs['keys'] or KeyTable()
        self.resources = kwargs['resources'] or ResourceTable()
        self.entities = kwargs['entities'] or EntityTable()
        self.objects = kwargs['objects'] or ObjectTable()
        self.object_keys = kwargs['object_keys'] or ObjectKeyTable()
        self.type_map = kwargs['type_map'] or get_type_map()

    @staticmethod
    def assert_external_resources_equal(left, right, check_dtype=True):
        """
        Compare that the keys, resources, entities, objects, and object_keys tables match

        :param left: ExternalResources object to compare with right
        :param right: ExternalResources object to compare with left
        :param check_dtype: Enforce strict checking of dtypes. Dtypes may be different
            for example for ids, where depending on how the data was saved
            ids may change from int64 to int32. (Default: True)
        :returns: The function returns True if all values match. If mismatches are found,
            AssertionError will be raised.
        :raises AssertionError: Raised if any differences are found. The function collects
            all differences into a single error so that the assertion will indicate
            all found differences.
        """
        errors = []
        try:
            pd.testing.assert_frame_equal(left.keys.to_dataframe(),
                                          right.keys.to_dataframe(),
                                          check_dtype=check_dtype)
        except AssertionError as e:
            errors.append(e)
        try:
            pd.testing.assert_frame_equal(left.objects.to_dataframe(),
                                          right.objects.to_dataframe(),
                                          check_dtype=check_dtype)
        except AssertionError as e:
            errors.append(e)
        try:
            pd.testing.assert_frame_equal(left.resources.to_dataframe(),
                                          right.resources.to_dataframe(),
                                          check_dtype=check_dtype)
        except AssertionError as e:
            errors.append(e)
        try:
            pd.testing.assert_frame_equal(left.entities.to_dataframe(),
                                          right.entities.to_dataframe(),
                                          check_dtype=check_dtype)
        except AssertionError as e:
            errors.append(e)
        try:
            pd.testing.assert_frame_equal(left.object_keys.to_dataframe(),
                                          right.object_keys.to_dataframe(),
                                          check_dtype=check_dtype)
        except AssertionError as e:
            errors.append(e)
        if len(errors) > 0:
            msg = ''.join(str(e)+"\n\n" for e in errors)
            raise AssertionError(msg)
        return True

    @docval({'name': 'key_name', 'type': str, 'doc': 'The name of the key to be added.'})
    def _add_key(self, **kwargs):
        """
        Add a key to be used for making references to external resources.

        It is possible to use the same *key_name* to refer to different resources so long as the *key_name* is not
        used within the same object, relative_path, and field. To do so, this method must be called for the
        two different resources.

        The returned Key objects must be managed by the caller so as to be appropriately passed to subsequent calls
        to methods for storing information about the different resources.
        """
        key = kwargs['key_name']
        return Key(key, table=self.keys)

    @docval({'name': 'key', 'type': (str, Key), 'doc': 'The key to associate the entity with.'},
            {'name': 'resources_idx', 'type': (int, Resource), 'doc': 'The id of the resource.'},
            {'name': 'entity_id', 'type': str, 'doc': 'The unique entity id.'},
            {'name': 'entity_uri', 'type': str, 'doc': 'The URI for the entity.'})
    def _add_entity(self, **kwargs):
        """
        Add an entity that will be referenced to using the given key.
        """
        key = kwargs['key']
        resources_idx = kwargs['resources_idx']
        entity_id = kwargs['entity_id']
        entity_uri = kwargs['entity_uri']
        if not isinstance(key, Key):
            key = self._add_key(key)
        resource_entity = Entity(key, resources_idx, entity_id, entity_uri, table=self.entities)
        return resource_entity

    @docval({'name': 'resource', 'type': str, 'doc': 'The name of the ontology resource.'},
            {'name': 'uri', 'type': str, 'doc': 'The URI associated with ontology resource.'})
    def _add_resource(self, **kwargs):
        """
        Add resource name and URI to ResourceTable that will be referenced by the ResourceTable idx.
        """
        resource_name = kwargs['resource']
        uri = kwargs['uri']
        resource = Resource(resource_name, uri, table=self.resources)
        return resource

    @docval({'name': 'container', 'type': (str, AbstractContainer),
             'doc': 'The Container/Data object to add or the object id of the Container/Data object to add.'},
            {'name': 'relative_path', 'type': str,
             'doc': ('The relative_path of the attribute of the object that uses ',
                     'an external resource reference key. Use an empty string if not applicable.')},
            {'name': 'field', 'type': str, 'default': '',
             'doc': ('The field of the compound data type using an external resource.')})
    def _add_object(self, **kwargs):
        """
        Add an object that references an external resource.
        """
        container, relative_path, field = popargs('container', 'relative_path', 'field', kwargs)
        if isinstance(container, AbstractContainer):
            container = container.object_id
        obj = Object(container, relative_path, field, table=self.objects)
        return obj

    @docval({'name': 'obj', 'type': (int, Object), 'doc': 'The Object that uses the Key.'},
            {'name': 'key', 'type': (int, Key), 'doc': 'The Key that the Object uses.'})
    def _add_object_key(self, **kwargs):
        """
        Specify that an object (i.e. container and relative_path) uses a key to reference
        an external resource.
        """
        obj, key = popargs('obj', 'key', kwargs)
        return ObjectKey(obj, key, table=self.object_keys)

    @docval({'name': 'container', 'type': (str, AbstractContainer),
             'doc': ('The Container/Data object that uses the key or '
                     'the object id for the Container/Data object that uses the key.')},
            {'name': 'relative_path', 'type': str,
             'doc': ('The relative_path of the attribute of the object that uses ',
                     'an external resource reference key. Use an empty string if not applicable.'),
             'default': ''},
            {'name': 'field', 'type': str, 'default': '',
             'doc': ('The field of the compound data type using an external resource.')},
            {'name': 'create', 'type': bool, 'default': True})
    def _check_object_field(self, container, relative_path, field, create):
        """
        Check if a container, relative path, and field have been added.

        The container can be either an object_id string or an AbstractContainer.

        If the container, relative_path, and field have not been added, add them
        and return the corresponding Object. Otherwise, just return the Object.
        """
        if isinstance(container, str):
            objecttable_idx = self.objects.which(object_id=container)
        else:
            objecttable_idx = self.objects.which(object_id=container.object_id)

        if len(objecttable_idx) > 0:
            relative_path_idx = self.objects.which(relative_path=relative_path)
            field_idx = self.objects.which(field=field)
            objecttable_idx = list(set(objecttable_idx) & set(relative_path_idx) & set(field_idx))

        if len(objecttable_idx) == 1:
            return self.objects.row[objecttable_idx[0]]
        elif len(objecttable_idx) == 0 and create:
            return self._add_object(container, relative_path, field)
        elif len(objecttable_idx) == 0 and not create:
            raise ValueError("Object not in Object Table.")
        else:
            raise ValueError("Found multiple instances of the same object id, relative path, "
                             "and field in objects table.")

    @docval({'name': 'key_name', 'type': str, 'doc': 'The name of the Key to get.'},
            {'name': 'container', 'type': (str, AbstractContainer), 'default': None,
             'doc': ('The Container/Data object that uses the key or '
                     'the object id for the Container/Data object that uses the key.')},
            {'name': 'relative_path', 'type': str,
             'doc': ('The relative_path of the attribute of the object that uses ',
                     'an external resource reference key. Use an empty string if not applicable.'),
             'default': ''},
            {'name': 'field', 'type': str, 'default': '',
             'doc': ('The field of the compound data type using an external resource.')})
    def get_key(self, **kwargs):
        """
        Return a Key or a list of Key objects that correspond to the given key.

        If container, relative_path, and field are provided, the Key that corresponds to the given name of the key
        for the given container, relative_path, and field is returned.
        """
        key_name, container, relative_path, field = popargs('key_name', 'container', 'relative_path', 'field', kwargs)
        key_idx_matches = self.keys.which(key=key_name)

        if container is not None:
            # if same key is used multiple times, determine
            # which instance based on the Container
            object_field = self._check_object_field(container, relative_path, field)
            for row_idx in self.object_keys.which(objects_idx=object_field.idx):
                key_idx = self.object_keys['keys_idx', row_idx]
                if key_idx in key_idx_matches:
                    return self.keys.row[key_idx]
            msg = ("No key '%s' found for container '%s', relative_path '%s', and field '%s'"
                   % (key_name, container, relative_path, field))
            raise ValueError(msg)
        else:
            if len(key_idx_matches) == 0:
                # the key has never been used before
                raise ValueError("key '%s' does not exist" % key_name)
            elif len(key_idx_matches) > 1:
                return [self.keys.row[i] for i in key_idx_matches]
            else:
                return self.keys.row[key_idx_matches[0]]

    @docval({'name': 'resource_name', 'type': str, 'doc': 'The name of the resource.'})
    def get_resource(self, **kwargs):
        """
        Retrieve resource object with the given resource_name.
        """
        resource_table_idx = self.resources.which(resource=kwargs['resource_name'])
        if len(resource_table_idx) == 0:
            # Resource hasn't been created
            msg = "No resource '%s' exists. Use _add_resource to create a new resource" % kwargs['resource_name']
            raise ValueError(msg)
        else:
            return self.resources.row[resource_table_idx[0]]

    @docval({'name': 'container', 'type': (str, AbstractContainer), 'default': None,
             'doc': ('The Container/Data object that uses the key or '
                     'the object_id for the Container/Data object that uses the key.')},
            {'name': 'attribute', 'type': str,
             'doc': 'The attribute of the container for the external reference.', 'default': None},
            {'name': 'field', 'type': str, 'default': '',
             'doc': ('The field of the compound data type using an external resource.')},
            {'name': 'key', 'type': (str, Key), 'default': None,
             'doc': 'The name of the key or the Key object from the KeyTable for the key to add a resource for.'},
            {'name': 'resources_idx', 'type': Resource, 'doc': 'The Resource from the ResourceTable.', 'default': None},
            {'name': 'resource_name', 'type': str, 'doc': 'The name of the resource to be created.', 'default': None},
            {'name': 'resource_uri', 'type': str, 'doc': 'The URI of the resource to be created.', 'default': None},
            {'name': 'entity_id', 'type': str, 'doc': 'The identifier for the entity at the resource.',
             'default': None},
            {'name': 'entity_uri', 'type': str, 'doc': 'The URI for the identifier at the resource.', 'default': None}
            )
    def add_ref(self, **kwargs):
        """
        Add information about an external reference used in this file.

        It is possible to use the same name of the key to refer to different resources
        so long as the name of the key is not used within the same object, relative_path, and
        field combination. This method does not support such functionality by default.
        """
        ###############################################################
        container = kwargs['container']
        attribute = kwargs['attribute']
        key = kwargs['key']
        field = kwargs['field']
        entity_id = kwargs['entity_id']
        entity_uri = kwargs['entity_uri']
        add_entity = False

        if attribute is None:  # Trivial Case
            relative_path = ''
            object_field = self._check_object_field(container, relative_path, field)
        else:  # DataType Attribute Case
            attribute_object = getattr(container, attribute)  # returns attribute object
            if isinstance(attribute_object, AbstractContainer):
                relative_path = ''
                object_field = self._check_object_field(attribute_object, relative_path, field)
            else:  # Non-DataType Attribute Case:
                obj_mapper = self.type_map.get_map(container)
                spec = obj_mapper.get_attr_spec(attr_name=attribute)
                parent_spec = spec.parent  # return the parent spec of the attribute
                if parent_spec.data_type is None:
                    while parent_spec.data_type is None:
                        parent_spec = parent_spec.parent  # find the closest parent with a data_type
                    parent_cls = self.type_map.get_dt_container_cls(data_type=parent_spec.data_type, autogen=False)
                    if isinstance(container, parent_cls):
                        parent_id = container.object_id
                        # We need to get the path of the spec for relative_path
                        absolute_path = spec.path
                        relative_path = re.sub("^.+?(?="+container.data_type+")", "", absolute_path)
                        object_field = self._check_object_field(parent_id, relative_path, field)
                    else:
                        msg = 'Container not the nearest data_type'
                        raise ValueError(msg)
                else:
                    parent_id = container.object_id  # container needs to be the parent
                    absolute_path = spec.path
                    relative_path = re.sub("^.+?(?="+container.data_type+")", "", absolute_path)
                    # this regex removes everything prior to the container on the absolute_path
                    object_field = self._check_object_field(parent_id, relative_path, field)

        if not isinstance(key, Key):
            key_idx_matches = self.keys.which(key=key)
        # if same key is used multiple times, determine
        # which instance based on the Container
            for row_idx in self.object_keys.which(objects_idx=object_field.idx):
                key_idx = self.object_keys['keys_idx', row_idx]
                if key_idx in key_idx_matches:
                    msg = "Use Key Object when referencing an existing (container, relative_path, key)"
                    raise ValueError(msg)

        if not isinstance(key, Key):
            key = self._add_key(key)
            self._add_object_key(object_field, key)

        if kwargs['resources_idx'] is not None and kwargs['resource_name'] is None and kwargs['resource_uri'] is None:
            resource_table_idx = kwargs['resources_idx']
        elif (
            kwargs['resources_idx'] is not None
            and (kwargs['resource_name'] is not None
                 or kwargs['resource_uri'] is not None)):
            msg = "Can't have resource_idx with resource_name or resource_uri."
            raise ValueError(msg)
        elif len(self.resources.which(resource=kwargs['resource_name'])) == 0:
            resource_name = kwargs['resource_name']
            resource_uri = kwargs['resource_uri']
            resource_table_idx = self._add_resource(resource_name, resource_uri)
        else:
            idx = self.resources.which(resource=kwargs['resource_name'])
            resource_table_idx = self.resources.row[idx[0]]

        if (resource_table_idx is not None and entity_id is not None and entity_uri is not None):
            add_entity = True
        elif not (resource_table_idx is None and entity_id is None and resource_uri is None):
            msg = ("Specify resource, entity_id, and entity_uri arguments."
                   "All three are required to create a reference")
            raise ValueError(msg)

        if add_entity:
            entity = self._add_entity(key, resource_table_idx, entity_id, entity_uri)

        return key, resource_table_idx, entity

    @docval({'name': 'container', 'type': (str, AbstractContainer),
             'doc': 'The Container/data object that is linked to resources/entities.'},
            {'name': 'relative_path', 'type': str,
             'doc': ('The relative_path of the attribute of the object that uses ',
                     'an external resource reference key. Use an empty string if not applicable.'),
             'default': ''},
            {'name': 'field', 'type': str, 'default': '',
             'doc': ('The field of the compound data type using an external resource.')})
    def get_object_resources(self, **kwargs):
        """
        Get all entities/resources associated with an object.
        """
        container = kwargs['container']
        relative_path = kwargs['relative_path']
        field = kwargs['field']

        keys = []
        entities = []
        object_field = self._check_object_field(container=container, relative_path=relative_path,
                                                field=field, create=False)
        # Find all keys associated with the object
        for row_idx in self.object_keys.which(objects_idx=object_field.idx):
            keys.append(self.object_keys['keys_idx', row_idx])
        # Find all the entities/resources for each key.
        for key_idx in keys:
            entity_idx = self.entities.which(keys_idx=key_idx)
            entities.append(list(self.entities.__getitem__(entity_idx[0])))
        df = pd.DataFrame(entities, columns=['keys_idx', 'resource_idx', 'entity_id', 'entity_uri'])
        return df

    @docval({'name': 'keys', 'type': (list, Key), 'default': None,
             'doc': 'The Key(s) to get external resource data for.'},
            rtype=pd.DataFrame, returns='a DataFrame with keys and external resource data')
    def get_keys(self, **kwargs):
        """
        Return a DataFrame with information about keys used to make references to external resources.
        The DataFrame will contain the following columns:
            - *key_name*:              the key that will be used for referencing an external resource
            - *resources_idx*:         the index for the resourcetable
            - *entity_id*:    the index for the entity at the external resource
            - *entity_uri*:   the URI for the entity at the external resource

        It is possible to use the same *key_name* to refer to different resources so long as the *key_name* is not
        used within the same object, relative_path, field. This method doesn't support such functionality by default. To
        select specific keys, use the *keys* argument to pass in the Key object(s) representing the desired keys. Note,
        if the same *key_name* is used more than once, multiple calls to this method with different Key objects will
        be required to keep the different instances separate. If a single call is made, it is left up to the caller to
        distinguish the different instances.
        """
        keys = popargs('keys', kwargs)
        if keys is None:
            keys = [self.keys.row[i] for i in range(len(self.keys))]
        else:
            if not isinstance(keys, list):
                keys = [keys]
        data = list()
        for key in keys:
            rsc_ids = self.entities.which(keys_idx=key.idx)
            for rsc_id in rsc_ids:
                rsc_row = self.entities.row[rsc_id].todict()
                rsc_row.pop('keys_idx')
                rsc_row['key_name'] = key.key
                data.append(rsc_row)
        return pd.DataFrame(data=data, columns=['key_name', 'resources_idx',
                                                'entity_id', 'entity_uri'])

    @docval({'name': 'use_categories', 'type': bool, 'default': False,
             'doc': 'Use a multi-index on the columns to indicate which category each column belongs to.'},
            rtype=pd.DataFrame, returns='A DataFrame with all data merged into a flat, denormalized table.')
    def to_dataframe(self, **kwargs):
        """
        Convert the data from the keys, resources, entities, objects, and object_keys tables
        to a single joint dataframe. I.e., here data is being denormalized, e.g., keys that
        are used across multiple enities or objects will duplicated across the corresponding
        rows.

        Returns: :py:class:`~pandas.DataFrame` with all data merged into a single, flat, denormalized table.

        """
        use_categories = popargs('use_categories', kwargs)
        # Step 1: Combine the entities, keys, and resources,table
        entities_df = self.entities.to_dataframe()
        # Map the keys to the entities by 1) convert to dataframe, 2) select rows based on the keys_idx
        # from the entities table, expanding the dataframe to have the same number of rows as the
        # entities, and 3) reset the index to avoid duplicate values in the index, which causes errors when merging
        keys_mapped_df = self.keys.to_dataframe().iloc[entities_df['keys_idx']].reset_index(drop=True)
        # Map the resources to entities using the same strategy as for the keys
        resources_mapped_df = self.resources.to_dataframe().iloc[entities_df['resources_idx']].reset_index(drop=True)
        # Merge the mapped keys and resources with the entities tables
        entities_df = pd.concat(objs=[entities_df, keys_mapped_df, resources_mapped_df],
                                axis=1, verify_integrity=False)
        # Add a column for the entity id (for consistency with the other tables and to facilitate query)
        entities_df['entities_idx'] = entities_df.index

        # Step 2: Combine the the object_keys and objects tables
        object_keys_df = self.object_keys.to_dataframe()
        objects_mapped_df = self.objects.to_dataframe().iloc[object_keys_df['objects_idx']].reset_index(drop=True)
        object_keys_df = pd.concat(objs=[object_keys_df, objects_mapped_df],
                                   axis=1,
                                   verify_integrity=False)

        # Step 3: merge the combined entities_df and object_keys_df DataFrames
        result_df = pd.concat(
            # Create for each row in the objects_keys table a DataFrame with all corresponding data from all tables
            objs=[pd.merge(
                    # Find all entities that correspond to the row i of the object_keys_table
                    entities_df[entities_df['keys_idx'] == object_keys_df['keys_idx'].iloc[i]].reset_index(drop=True),
                    # Get a DataFrame for row i of the objects_keys_table
                    object_keys_df.iloc[[i, ]],
                    # Merge the entities and object_keys on the keys_idx column so that the values from the single
                    # object_keys_table row are copied across all corresponding rows in the entities table
                    on='keys_idx')
                  for i in range(len(object_keys_df))],
            # Concatenate the rows of the objs
            axis=0,
            verify_integrity=False)

        # Step 4: Clean up the index and sort columns by table type and name
        result_df.reset_index(inplace=True, drop=True)
        column_labels = [('objects', 'objects_idx'), ('objects', 'object_id'),
                         ('objects', 'relative_path'), ('objects', 'field'),
                         ('keys', 'keys_idx'), ('keys', 'key'),
                         ('resources', 'resources_idx'), ('resources', 'resource'), ('resources', 'resource_uri'),
                         ('entities', 'entities_idx'), ('entities', 'entity_id'), ('entities', 'entity_uri')]
        # sort the columns based on our custom order
        result_df = result_df.reindex(labels=[c[1] for c in column_labels],
                                      axis=1)
        # Add the categories if requested
        if use_categories:
            result_df.columns = pd.MultiIndex.from_tuples(column_labels)
        # return the result
        return result_df

    @docval({'name': 'db_file', 'type': str, 'doc': 'Name of the SQLite database file'})
    def to_sqlite(self, db_file):
        """
        Save the keys, resources, entities, objects, and object_keys tables using sqlite3 to the given db_file.

        The function will first create the tables (if they do not already exist) and then
        add the data from this ExternalResource object to the database. If the database file already
        exists, then the data will be appended as rows to the existing database tables.

        Note, the index values of foreign keys (e.g., keys_idx, objects_idx, resources_idx) in the tables
        will not match between the ExternalResources here and the exported database, but they are adjusted
        automatically here, to ensure the foreign keys point to the correct rows in the exported database.
        This is because: 1) ExternalResources uses 0-based indexing for foreign keys, whereas SQLite uses
        1-based indexing and 2) if data is appended to existing tables then a corresponding additional
        offset must be applied to the relevant foreign keys.

        :raises: The function will raise errors if connection to the database fails. If
            the given db_file already exists, then there is also the possibility that
            certain updates may result in errors if there are collisions between the
            new and existing data.
        """
        import sqlite3
        # connect to the database
        connection = sqlite3.connect(db_file)
        cursor = connection.cursor()
        # sql calls to setup the tables
        sql_create_keys_table = """ CREATE TABLE IF NOT EXISTS keys (
                                        id integer PRIMARY KEY,
                                        key text NOT NULL
                                    ); """
        sql_create_objects_table = """ CREATE TABLE IF NOT EXISTS objects (
                                            id integer PRIMARY KEY,
                                            object_id text NOT NULL,
                                            relative_path text NOT NULL,
                                            field text
                                       ); """
        sql_create_resources_table = """ CREATE TABLE IF NOT EXISTS resources (
                                             id integer PRIMARY KEY,
                                             resource text NOT NULL,
                                             resource_uri text NOT NULL
                                        ); """
        sql_create_object_keys_table = """ CREATE TABLE IF NOT EXISTS object_keys (
                                               id integer PRIMARY KEY,
                                               objects_idx int NOT NULL,
                                               keys_idx int NOT NULL,
                                               FOREIGN KEY (objects_idx) REFERENCES objects (id),
                                               FOREIGN KEY (keys_idx) REFERENCES keys (id)
                                        ); """
        sql_create_entities_table = """ CREATE TABLE IF NOT EXISTS entities (
                                             id integer PRIMARY KEY,
                                             keys_idx int NOT NULL,
                                             resources_idx int NOT NULL,
                                             entity_id text NOT NULL,
                                             entity_uri text NOT NULL,
                                             FOREIGN KEY (keys_idx) REFERENCES keys (id),
                                             FOREIGN KEY (resources_idx) REFERENCES resources (id)
                                        ); """
        # execute setting up the tables
        cursor.execute(sql_create_keys_table)
        cursor.execute(sql_create_objects_table)
        cursor.execute(sql_create_resources_table)
        cursor.execute(sql_create_object_keys_table)
        cursor.execute(sql_create_entities_table)

        # NOTE: sqlite uses a 1-based row-index so we need to update all foreign key columns accordingly
        # NOTE: If we are adding to an existing sqlite database then we need to also adjust for he number of rows
        keys_offset = len(cursor.execute('select * from keys;').fetchall()) + 1
        objects_offset = len(cursor.execute('select * from objects;').fetchall()) + 1
        resources_offset = len(cursor.execute('select * from resources;').fetchall()) + 1

        # populate the tables and fix foreign keys during insert
        cursor.executemany(" INSERT INTO keys(key) VALUES(?) ", self.keys[:])
        connection.commit()
        cursor.executemany(" INSERT INTO objects(object_id, relative_path, field) VALUES(?, ?, ?) ", self.objects[:])
        connection.commit()
        cursor.executemany(" INSERT INTO resources(resource, resource_uri) VALUES(?, ?) ", self.resources[:])
        connection.commit()
        cursor.executemany(
            " INSERT INTO object_keys(objects_idx, keys_idx) VALUES(?+%i, ?+%i) " % (objects_offset, keys_offset),
            self.object_keys[:])
        connection.commit()
        cursor.executemany(
            " INSERT INTO entities(keys_idx, resources_idx, entity_id, entity_uri) VALUES(?+%i, ?+%i, ?, ?) "
            % (keys_offset, resources_offset),
            self.entities[:])
        connection.commit()
        connection.close()

    @docval({'name': 'path', 'type': str, 'doc': 'path of the tsv file to write'})
    def to_tsv(self, **kwargs):
        """
        Write ExternalResources as a single, flat table to TSV
        Internally, the function uses :py:meth:`pandas.DataFrame.to_csv`. Pandas can
        infer compression based on the filename, i.e., by changing the file extension to
        ‘.gz’, ‘.bz2’, ‘.zip’, ‘.xz’, or ‘.zst’ we can write compressed files.
        The TSV is formatted as follows: 1) line one indicates for each column the name of the table
        the column belongs to, 2) line two is the name of the column within the table, 3) subsequent
        lines are each a row in the flattened ExternalResources table. The first column is the
        row id in the flattened table and does not have a label, i.e., the first and second
        row will start with a tab character, and subseqent rows are numbered sequentially 1,2,3,... .
        For example:

        .. code-block::
            :linenos:

            \tobjects\tobjects\tobjects\tobjects\tkeys\tkeys\tresources\tresources\tresources\tentities\tentities\tentities
            \tobjects_idx\tobject_id\trelative_path\tfield\tkeys_idx\tkey\tresources_idx\tresource\tresource_uri\tentities_idx\tentity_id\tentity_uri
            0\t0\t1fc87200-e91e-45b3-978c-6d295af144c3\t\tspecies\t0\tMus musculus\t0\tNCBI_Taxonomy\thttps://www.ncbi.nlm.nih.gov/taxonomy\t0\tNCBI:txid10090\thttps://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?id=10090
            1\t0\t9bf0c58e-09dc-4457-a652-94065b112c41\t\tspecies\t1\tHomo sapiens\t0\tNCBI_Taxonomy\thttps://www.ncbi.nlm.nih.gov/taxonomy\t1\tNCBI:txid9606\thttps://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?id=9606

        See also :py:meth:`~hdmf.common.resources.ExternalResources.from_tsv`
        """  # noqa: E501
        path = popargs('path', kwargs)
        df = self.to_dataframe(use_categories=True)
        df.to_csv(path, sep='\t')

    @classmethod
    @docval({'name': 'path', 'type': str, 'doc': 'path of the tsv file to read'},
            returns="ExternalResources loaded from TSV", rtype="ExternalResources")
    def from_tsv(cls, **kwargs):
        """
        Read ExternalResources from a flat tsv file
        Formatting of the TSV file is assumed to be consistent with the format
        generated by :py:meth:`~hdmf.common.resources.ExternalResources.to_tsv`.
        The function attempts to validate that the data in the TSV is consistent
        and parses the data from the denormalized table in the TSV to the
        normalized linked table structure used by ExternalResources.
        Currently the checks focus on ensuring that row id links between tables are valid.
        Inconsistencies in other (non-index) fields (e.g., when two rows with the same resource_idx
        have different resource_uri values) are not checked and will be ignored. In this case, the value
        from the first row that contains the corresponding entry will be kept.

        .. note::
           Since TSV files may be edited by hand or other applications, it is possible that data
           in the TSV may be inconsistent. E.g., object_idx may be missing if rows were removed
           and ids not updated. Also since the TSV is flattened into a single denormalized table
           (i.e., data are stored with duplication, rather than normalized across several tables),
           it is possible that values may be inconsistent if edited outside. E.g., we may have
           objects with the same index (object_idx) but different object_id, relative_path, or field
           values. While flat TSVs are sometimes preferred for ease of sharing, editing
           the TSV without using the :py:meth:`~hdmf.common.resources.ExternalResources` class
           should be done with great care!
        """
        def check_idx(idx_arr, name):
            """Check that indices are consecutively numbered without missing values"""
            idx_diff = np.diff(idx_arr)
            if np.any(idx_diff != 1):
                missing_idx = [i for i in range(np.max(idx_arr)) if i not in idx_arr]
                msg = "Missing %s entries %s" % (name, str(missing_idx))
                raise ValueError(msg)

        path = popargs('path', kwargs)
        df = pd.read_csv(path, header=[0, 1], sep='\t').replace(np.nan, '')
        # Construct the ExternalResources
        er = ExternalResources(name="external_resources")

        # Retrieve all the objects
        ob_idx, ob_rows = np.unique(df[('objects', 'objects_idx')], return_index=True)
        # Sort objects based on their index
        ob_order = np.argsort(ob_idx)
        ob_idx = ob_idx[ob_order]
        ob_rows = ob_rows[ob_order]
        # Check that objects are consecutively numbered
        check_idx(idx_arr=ob_idx, name='objects_idx')
        # Add the objects to the Object table
        ob_ids = df[('objects', 'object_id')].iloc[ob_rows]
        ob_relpaths = df[('objects', 'relative_path')].iloc[ob_rows]
        ob_fields = df[('objects', 'field')].iloc[ob_rows]
        for ob in zip(ob_ids, ob_relpaths, ob_fields):
            er._add_object(container=ob[0], relative_path=ob[1], field=ob[2])

        # Retrieve all keys
        keys_idx, keys_rows = np.unique(df[('keys', 'keys_idx')], return_index=True)
        # Sort keys based on their index
        keys_order = np.argsort(keys_idx)
        keys_idx = keys_idx[keys_order]
        keys_rows = keys_rows[keys_order]
        # Check that keys are consecutively numbered
        check_idx(idx_arr=keys_idx, name='keys_idx')
        # Add the keys to the Keys table
        keys_key = df[('keys', 'key')].iloc[keys_rows]
        all_added_keys = [er._add_key(k) for k in keys_key]

        # Add all the object keys to the ObjectKeys table. A single key may be assigned to multiple
        # objects. As such it is not sufficient to iterate over the unique ob_rows with the unique
        # objects, but we need to find all unique (objects_idx, keys_idx) combinations.
        ob_keys_idx = np.unique(df[[('objects', 'objects_idx'), ('keys', 'keys_idx')]], axis=0)
        for obk in ob_keys_idx:
            er._add_object_key(obj=obk[0], key=obk[1])

        # Retrieve all resources
        resources_idx, resources_rows = np.unique(df[('resources', 'resources_idx')], return_index=True)
        # Sort resources based on their index
        resources_order = np.argsort(resources_idx)
        resources_idx = resources_idx[resources_order]
        resources_rows = resources_rows[resources_order]
        # Check that resources are consecutively numbered
        check_idx(idx_arr=resources_idx, name='resources_idx')
        # Add the resources to the Resources table
        resources_resource = df[('resources', 'resource')].iloc[resources_rows]
        resources_uri = df[('resources', 'resource_uri')].iloc[resources_rows]
        for r in zip(resources_resource, resources_uri):
            er._add_resource(resource=r[0], uri=r[1])

        # Retrieve all entities
        entities_idx, entities_rows = np.unique(df[('entities', 'entities_idx')], return_index=True)
        # Sort entities based on their index
        entities_order = np.argsort(entities_idx)
        entities_idx = entities_idx[entities_order]
        entities_rows = entities_rows[entities_order]
        # Check that entities are consecutively numbered
        check_idx(idx_arr=entities_idx, name='entities_idx')
        # Add the entities ot the Resources table
        entities_id = df[('entities', 'entity_id')].iloc[entities_rows]
        entities_uri = df[('entities', 'entity_uri')].iloc[entities_rows]
        entities_keys = np.array(all_added_keys)[df[('keys', 'keys_idx')].iloc[entities_rows]]
        entities_resources_idx = df[('resources', 'resources_idx')].iloc[entities_rows]
        for e in zip(entities_keys, entities_resources_idx, entities_id, entities_uri):
            er._add_entity(key=e[0], resources_idx=e[1], entity_id=e[2], entity_uri=e[3])

        # Return the reconstructed ExternalResources
        return er



from abc import ABCMeta, abstractmethod

import numpy as np
from hdmf import Container, Data
from hdmf.build import ObjectMapper, BuildManager, TypeMap, GroupBuilder, DatasetBuilder
from hdmf.build.warnings import DtypeConversionWarning
from hdmf.spec import GroupSpec, AttributeSpec, DatasetSpec, SpecCatalog, SpecNamespace, NamespaceCatalog, Spec
from hdmf.testing import TestCase
from hdmf.utils import docval, getargs

from tests.unit.utils import CORE_NAMESPACE


# TODO: test build of extended group/dataset that modifies an attribute dtype (commented out below), shape, value, etc.
# by restriction. also check that attributes cannot be deleted or scope expanded.
# TODO: test build of extended dataset that modifies shape by restriction.

class Bar(Container):

    @docval({'name': 'name', 'type': str, 'doc': 'the name of this Bar'},
            {'name': 'attr1', 'type': str, 'doc': 'a string attribute'},
            {'name': 'attr2', 'type': 'int', 'doc': 'an int attribute', 'default': None},
            {'name': 'ext_attr', 'type': bool, 'doc': 'a boolean attribute', 'default': True})
    def __init__(self, **kwargs):
        name, attr1, attr2, ext_attr = getargs('name', 'attr1', 'attr2', 'ext_attr', kwargs)
        super().__init__(name=name)
        self.__attr1 = attr1
        self.__attr2 = attr2
        self.__ext_attr = kwargs['ext_attr']

    @property
    def data_type(self):
        return 'Bar'

    @property
    def attr1(self):
        return self.__attr1

    @property
    def attr2(self):
        return self.__attr2

    @property
    def ext_attr(self):
        return self.__ext_attr


class BarHolder(Container):

    @docval({'name': 'name', 'type': str, 'doc': 'the name of this BarHolder'},
            {'name': 'bars', 'type': ('data', 'array_data'), 'doc': 'bars', 'default': list()})
    def __init__(self, **kwargs):
        name, bars = getargs('name', 'bars', kwargs)
        super().__init__(name=name)
        self.__bars = bars
        for b in bars:
            if b is not None and b.parent is None:
                b.parent = self

    @property
    def data_type(self):
        return 'BarHolder'

    @property
    def bars(self):
        return self.__bars


class ExtBarMapper(ObjectMapper):

    @docval({"name": "spec", "type": Spec, "doc": "the spec to get the attribute value for"},
            {"name": "container", "type": Bar, "doc": "the container to get the attribute value from"},
            {"name": "manager", "type": BuildManager, "doc": "the BuildManager used for managing this build"},
            returns='the value of the attribute')
    def get_attr_value(self, **kwargs):
        ''' Get the value of the attribute corresponding to this spec from the given container '''
        spec, container, manager = getargs('spec', 'container', 'manager', kwargs)
        # handle custom mapping of field 'ext_attr' within container BarHolder/Bar -> spec BarHolder/Bar.ext_attr
        if isinstance(container.parent, BarHolder):
            if spec.name == 'ext_attr':
                return container.ext_attr
        return super().get_attr_value(**kwargs)


class BuildGroupExtAttrsMixin(TestCase, metaclass=ABCMeta):

    def setUp(self):
        self.setUpBarSpec()
        self.setUpBarHolderSpec()
        spec_catalog = SpecCatalog()
        spec_catalog.register_spec(self.bar_spec, 'test.yaml')
        spec_catalog.register_spec(self.bar_holder_spec, 'test.yaml')
        namespace = SpecNamespace(
            doc='a test namespace',
            name=CORE_NAMESPACE,
            schema=[{'source': 'test.yaml'}],
            version='0.1.0',
            catalog=spec_catalog
        )
        namespace_catalog = NamespaceCatalog()
        namespace_catalog.add_namespace(CORE_NAMESPACE, namespace)
        type_map = TypeMap(namespace_catalog)
        type_map.register_container_type(CORE_NAMESPACE, 'Bar', Bar)
        type_map.register_container_type(CORE_NAMESPACE, 'BarHolder', BarHolder)
        type_map.register_map(Bar, ExtBarMapper)
        type_map.register_map(BarHolder, ObjectMapper)
        self.manager = BuildManager(type_map)

    def setUpBarSpec(self):
        attr1_attr = AttributeSpec(
            name='attr1',
            dtype='text',
            doc='an example string attribute',
        )
        attr2_attr = AttributeSpec(
            name='attr2',
            dtype='int',
            doc='an example int attribute',
        )
        self.bar_spec = GroupSpec(
            doc='A test group specification with a data type',
            data_type_def='Bar',
            attributes=[attr1_attr, attr2_attr],
        )

    @abstractmethod
    def setUpBarHolderSpec(self):
        pass


class TestBuildGroupAddedAttr(BuildGroupExtAttrsMixin, TestCase):
    """
    If the spec defines a group data_type A (Bar) using 'data_type_def' and defines another data_type B (BarHolder)
    that includes A using 'data_type_inc', then the included A spec is an extended (or refined) spec of A - call it A'.
    The spec of A' can refine or add attributes to the spec of A. This test ensures that *added attributes* in A' are
    handled properly.
    """

    def setUpBarHolderSpec(self):
        ext_attr = AttributeSpec(
            name='ext_attr',
            dtype='bool',
            doc='A boolean attribute',
        )
        bar_ext_no_name_spec = GroupSpec(
            doc='A Bar extended with attribute ext_attr',
            data_type_inc='Bar',
            quantity='*',
            attributes=[ext_attr],
        )
        self.bar_holder_spec = GroupSpec(
            doc='A container of multiple extended Bar objects',
            data_type_def='BarHolder',
            groups=[bar_ext_no_name_spec],
        )

    def test_build_added_attr(self):
        """
        Test build of BarHolder which can contain multiple extended Bar objects, which have a new attribute.
        """
        ext_bar_inst = Bar(
            name='my_bar',
            attr1='a string',
            attr2=10,
            ext_attr=False,
        )
        bar_holder_inst = BarHolder(
            name='my_bar_holder',
            bars=[ext_bar_inst],
        )

        expected_inner = GroupBuilder(
            name='my_bar',
            attributes={
                'attr1': 'a string',
                'attr2': 10,
                'data_type': 'Bar',
                'ext_attr': False,
                'namespace': CORE_NAMESPACE,
                'object_id': ext_bar_inst.object_id,
            },
        )
        expected = GroupBuilder(
            name='my_bar_holder',
            groups={'my_bar': expected_inner},
            attributes={
                'data_type': 'BarHolder',
                'namespace': CORE_NAMESPACE,
                'object_id': bar_holder_inst.object_id,
            },
        )

        # the object mapper automatically maps the spec of extended Bars to the 'BarMapper.bars' field
        builder = self.manager.build(bar_holder_inst, source='test.h5')
        self.assertDictEqual(builder, expected)


class TestBuildGroupRefinedAttr(BuildGroupExtAttrsMixin, TestCase):
    """
    If the spec defines a group data_type A (Bar) using 'data_type_def' and defines another data_type B (BarHolder)
    that includes A using 'data_type_inc', then the included A spec is an extended (or refined) spec of A - call it A'.
    The spec of A' can refine or add attributes to the spec of A. This test ensures that *refine attributes* in A' are
    handled properly.
    """

    def setUpBarHolderSpec(self):
        int_attr2 = AttributeSpec(
            name='attr2',
            dtype='int64',
            doc='Refine Bar spec from int to int64',
        )
        bar_ext_no_name_spec = GroupSpec(
            doc='A Bar extended with modified attribute attr2',
            data_type_inc='Bar',
            quantity='*',
            attributes=[int_attr2],
        )
        self.bar_holder_spec = GroupSpec(
            doc='A container of multiple extended Bar objects',
            data_type_def='BarHolder',
            groups=[bar_ext_no_name_spec],
        )

    def test_build_refined_attr(self):
        """
        Test build of BarHolder which can contain multiple extended Bar objects, which have a modified attr2.
        """
        ext_bar_inst = Bar(
            name='my_bar',
            attr1='a string',
            attr2=np.int64(10),
        )
        bar_holder_inst = BarHolder(
            name='my_bar_holder',
            bars=[ext_bar_inst],
        )

        expected_inner = GroupBuilder(
            name='my_bar',
            attributes={
                'attr1': 'a string',
                'attr2': np.int64(10),
                'data_type': 'Bar',
                'namespace': CORE_NAMESPACE,
                'object_id': ext_bar_inst.object_id,
            }
        )
        expected = GroupBuilder(
            name='my_bar_holder',
            groups={'my_bar': expected_inner},
            attributes={
                'data_type': 'BarHolder',
                'namespace': CORE_NAMESPACE,
                'object_id': bar_holder_inst.object_id,
            },
        )

        # the object mapper automatically maps the spec of extended Bars to the 'BarMapper.bars' field
        builder = self.manager.build(bar_holder_inst, source='test.h5')
        self.assertDictEqual(builder, expected)

    # def test_build_refined_attr_wrong_type(self):
    #     """
    #     Test build of BarHolder which contains a Bar that has the wrong dtype for an attr.
    #     """
    #     ext_bar_inst = Bar(
    #         name='my_bar',
    #         attr1='a string',
    #         attr2=10,  # spec specifies attr2 should be an int64 for Bars within BarHolder
    #     )
    #     bar_holder_inst = BarHolder(
    #         name='my_bar_holder',
    #         bars=[ext_bar_inst],
    #     )
    #
    #     expected_inner = GroupBuilder(
    #         name='my_bar',
    #         attributes={
    #             'attr1': 'a string',
    #             'attr2': np.int64(10),
    #             'data_type': 'Bar',
    #             'namespace': CORE_NAMESPACE,
    #             'object_id': ext_bar_inst.object_id,
    #         }
    #     )
    #     expected = GroupBuilder(
    #         name='my_bar_holder',
    #         groups={'my_bar': expected_inner},
    #         attributes={
    #             'data_type': 'BarHolder',
    #             'namespace': CORE_NAMESPACE,
    #             'object_id': bar_holder_inst.object_id,
    #         },
    #     )
    #
    #     # the object mapper automatically maps the spec of extended Bars to the 'BarMapper.bars' field
    #
    #     # TODO build should raise a conversion warning for converting 10 (int32) to np.int64
    #     builder = self.manager.build(bar_holder_inst, source='test.h5')
    #     self.assertDictEqual(builder, expected)


class BarData(Data):

    @docval({'name': 'name', 'type': str, 'doc': 'the name of this BarData'},
            {'name': 'data', 'type': ('data', 'array_data'), 'doc': 'the data'},
            {'name': 'attr1', 'type': str, 'doc': 'a string attribute'},
            {'name': 'attr2', 'type': 'int', 'doc': 'an int attribute', 'default': None},
            {'name': 'ext_attr', 'type': bool, 'doc': 'a boolean attribute', 'default': True})
    def __init__(self, **kwargs):
        name, data, attr1, attr2, ext_attr = getargs('name', 'data', 'attr1', 'attr2', 'ext_attr', kwargs)
        super().__init__(name=name, data=data)
        self.__attr1 = attr1
        self.__attr2 = attr2
        self.__ext_attr = kwargs['ext_attr']

    @property
    def data_type(self):
        return 'BarData'

    @property
    def attr1(self):
        return self.__attr1

    @property
    def attr2(self):
        return self.__attr2

    @property
    def ext_attr(self):
        return self.__ext_attr


class BarDataHolder(Container):

    @docval({'name': 'name', 'type': str, 'doc': 'the name of this BarDataHolder'},
            {'name': 'bar_datas', 'type': ('data', 'array_data'), 'doc': 'bar_datas', 'default': list()})
    def __init__(self, **kwargs):
        name, bar_datas = getargs('name', 'bar_datas', kwargs)
        super().__init__(name=name)
        self.__bar_datas = bar_datas
        for b in bar_datas:
            if b is not None and b.parent is None:
                b.parent = self

    @property
    def data_type(self):
        return 'BarDataHolder'

    @property
    def bar_datas(self):
        return self.__bar_datas


class ExtBarDataMapper(ObjectMapper):

    @docval({"name": "spec", "type": Spec, "doc": "the spec to get the attribute value for"},
            {"name": "container", "type": BarData, "doc": "the container to get the attribute value from"},
            {"name": "manager", "type": BuildManager, "doc": "the BuildManager used for managing this build"},
            returns='the value of the attribute')
    def get_attr_value(self, **kwargs):
        ''' Get the value of the attribute corresponding to this spec from the given container '''
        spec, container, manager = getargs('spec', 'container', 'manager', kwargs)
        # handle custom mapping of field 'ext_attr' within container
        # BardataHolder/BarData -> spec BarDataHolder/BarData.ext_attr
        if isinstance(container.parent, BarDataHolder):
            if spec.name == 'ext_attr':
                return container.ext_attr
        return super().get_attr_value(**kwargs)


class BuildDatasetExtAttrsMixin(TestCase, metaclass=ABCMeta):

    def setUp(self):
        self.set_up_specs()
        spec_catalog = SpecCatalog()
        spec_catalog.register_spec(self.bar_data_spec, 'test.yaml')
        spec_catalog.register_spec(self.bar_data_holder_spec, 'test.yaml')
        namespace = SpecNamespace(
            doc='a test namespace',
            name=CORE_NAMESPACE,
            schema=[{'source': 'test.yaml'}],
            version='0.1.0',
            catalog=spec_catalog
        )
        namespace_catalog = NamespaceCatalog()
        namespace_catalog.add_namespace(CORE_NAMESPACE, namespace)
        type_map = TypeMap(namespace_catalog)
        type_map.register_container_type(CORE_NAMESPACE, 'BarData', BarData)
        type_map.register_container_type(CORE_NAMESPACE, 'BarDataHolder', BarDataHolder)
        type_map.register_map(BarData, ExtBarDataMapper)
        type_map.register_map(BarDataHolder, ObjectMapper)
        self.manager = BuildManager(type_map)

    def set_up_specs(self):
        attr1_attr = AttributeSpec(
            name='attr1',
            dtype='text',
            doc='an example string attribute',
        )
        attr2_attr = AttributeSpec(
            name='attr2',
            dtype='int',
            doc='an example int attribute',
        )
        self.bar_data_spec = DatasetSpec(
            doc='A test dataset specification with a data type',
            data_type_def='BarData',
            dtype='int',
            shape=[[None], [None, None]],
            attributes=[attr1_attr, attr2_attr],
        )
        self.bar_data_holder_spec = GroupSpec(
            doc='A container of multiple extended BarData objects',
            data_type_def='BarDataHolder',
            datasets=[self.get_refined_bar_data_spec()],
        )

    @abstractmethod
    def get_refined_bar_data_spec(self):
        pass


class TestBuildDatasetAddedAttrs(BuildDatasetExtAttrsMixin, TestCase):
    """
    If the spec defines a dataset data_type A (BarData) using 'data_type_def' and defines another data_type B
    (BarHolder) that includes A using 'data_type_inc', then the included A spec is an extended (or refined) spec of A -
    call it A'. The spec of A' can refine or add attributes, refine the dtype, refine the shape, or set a fixed value
    to the spec of A. This test ensures that *added attributes* in A' are handled properly. This is similar to how the
    spec for a subtype of DynamicTable can contain a VectorData that has a new attribute.
    """

    def get_refined_bar_data_spec(self):
        ext_attr = AttributeSpec(
            name='ext_attr',
            dtype='bool',
            doc='A boolean attribute',
        )
        refined_spec = DatasetSpec(
            doc='A BarData extended with attribute ext_attr',
            data_type_inc='BarData',
            quantity='*',
            attributes=[ext_attr],
        )
        return refined_spec

    def test_build_added_attr(self):
        """
        Test build of BarHolder which can contain multiple extended BarData objects, which have a new attribute.
        """
        ext_bar_data_inst = BarData(
            name='my_bar',
            data=list(range(10)),
            attr1='a string',
            attr2=10,
            ext_attr=False,
        )
        bar_data_holder_inst = BarDataHolder(
            name='my_bar_holder',
            bar_datas=[ext_bar_data_inst],
        )

        expected_inner = DatasetBuilder(
            name='my_bar',
            data=list(range(10)),
            attributes={
                'attr1': 'a string',
                'attr2': 10,
                'data_type': 'BarData',
                'ext_attr': False,
                'namespace': CORE_NAMESPACE,
                'object_id': ext_bar_data_inst.object_id,
            },
        )
        expected = GroupBuilder(
            name='my_bar_holder',
            datasets={'my_bar': expected_inner},
            attributes={
                'data_type': 'BarDataHolder',
                'namespace': CORE_NAMESPACE,
                'object_id': bar_data_holder_inst.object_id,
            },
        )

        # the object mapper automatically maps the spec of extended Bars to the 'BarMapper.bars' field
        builder = self.manager.build(bar_data_holder_inst, source='test.h5')
        self.assertDictEqual(builder, expected)


class TestBuildDatasetRefinedDtype(BuildDatasetExtAttrsMixin, TestCase):
    """
    If the spec defines a dataset data_type A (BarData) using 'data_type_def' and defines another data_type B
    (BarHolder) that includes A using 'data_type_inc', then the included A spec is an extended (or refined) spec of A -
    call it A'. The spec of A' can refine or add attributes, refine the dtype, refine the shape, or set a fixed value
    to the spec of A. This test ensures that if A' refines the dtype of A, the build process uses the correct dtype for
    conversion.
    """

    def get_refined_bar_data_spec(self):
        refined_spec = DatasetSpec(
            doc='A BarData with refined int64 dtype',
            data_type_inc='BarData',
            dtype='int64',
            quantity='*',
        )
        return refined_spec

    def test_build_refined_dtype_convert(self):
        """
        Test build of BarDataHolder which contains a BarData with data that needs to be converted to the refined dtype.
        """
        ext_bar_data_inst = BarData(
            name='my_bar',
            data=np.array([1, 2], dtype=np.int32),  # the refined spec says data should be int64s
            attr1='a string',
            attr2=10,
        )
        bar_data_holder_inst = BarDataHolder(
            name='my_bar_holder',
            bar_datas=[ext_bar_data_inst],
        )

        expected_inner = DatasetBuilder(
            name='my_bar',
            data=np.array([1, 2], dtype=np.int64),  # the objectmapper should convert the given data to int64s
            attributes={
                'attr1': 'a string',
                'attr2': 10,
                'data_type': 'BarData',
                'namespace': CORE_NAMESPACE,
                'object_id': ext_bar_data_inst.object_id,
            },
        )
        expected = GroupBuilder(
            name='my_bar_holder',
            datasets={'my_bar': expected_inner},
            attributes={
                'data_type': 'BarDataHolder',
                'namespace': CORE_NAMESPACE,
                'object_id': bar_data_holder_inst.object_id,
            },
        )

        # the object mapper automatically maps the spec of extended Bars to the 'BarMapper.bars' field
        msg = ("Spec 'BarDataHolder/BarData': Value with data type int32 is being converted to data type int64 "
               "as specified.")
        with self.assertWarnsWith(DtypeConversionWarning, msg):
            builder = self.manager.build(bar_data_holder_inst, source='test.h5')
        np.testing.assert_array_equal(builder.datasets['my_bar'].data, expected.datasets['my_bar'].data)
        self.assertEqual(builder.datasets['my_bar'].data.dtype, np.int64)


class TestBuildDatasetNotRefinedDtype(BuildDatasetExtAttrsMixin, TestCase):
    """
    If the spec defines a dataset data_type A (BarData) using 'data_type_def' and defines another data_type B
    (BarHolder) that includes A using 'data_type_inc', then the included A spec is an extended (or refined) spec of A -
    call it A'. The spec of A' can refine or add attributes, refine the dtype, refine the shape, or set a fixed value
    to the spec of A. This test ensures that if A' does not refine the dtype of A, the build process uses the correct
    dtype for conversion.
    """

    def get_refined_bar_data_spec(self):
        refined_spec = DatasetSpec(
            doc='A BarData',
            data_type_inc='BarData',
            quantity='*',
        )
        return refined_spec

    def test_build_correct_dtype(self):
        """
        Test build of BarDataHolder which contains a BarData.
        """
        ext_bar_data_inst = BarData(
            name='my_bar',
            data=[1, 2],
            attr1='a string',
            attr2=10,
        )
        bar_data_holder_inst = BarDataHolder(
            name='my_bar_holder',
            bar_datas=[ext_bar_data_inst],
        )

        expected_inner = DatasetBuilder(
            name='my_bar',
            data=[1, 2],
            attributes={
                'attr1': 'a string',
                'attr2': 10,
                'data_type': 'BarData',
                'namespace': CORE_NAMESPACE,
                'object_id': ext_bar_data_inst.object_id,
            },
        )
        expected = GroupBuilder(
            name='my_bar_holder',
            datasets={'my_bar': expected_inner},
            attributes={
                'data_type': 'BarDataHolder',
                'namespace': CORE_NAMESPACE,
                'object_id': bar_data_holder_inst.object_id,
            },
        )

        # the object mapper automatically maps the spec of extended Bars to the 'BarMapper.bars' field
        builder = self.manager.build(bar_data_holder_inst, source='test.h5')
        self.assertDictEqual(builder, expected)

    def test_build_incorrect_dtype(self):
        """
        Test build of BarDataHolder which contains a BarData
        """
        ext_bar_data_inst = BarData(
            name='my_bar',
            data=['a', 'b'],
            attr1='a string',
            attr2=10,
        )
        bar_data_holder_inst = BarDataHolder(
            name='my_bar_holder',
            bar_datas=[ext_bar_data_inst],
        )

        # the object mapper automatically maps the spec of extended Bars to the 'BarMapper.bars' field
        msg = "could not resolve dtype for BarData 'my_bar'"
        with self.assertRaisesWith(Exception, msg):
            self.manager.build(bar_data_holder_inst, source='test.h5')



import numpy as np
from pandas.testing import assert_frame_equal
import warnings

from hdmf.backends.hdf5 import HDF5IO
from hdmf.common import DynamicTable, VectorData, get_manager, AlignedDynamicTable, DynamicTableRegion
from hdmf.testing import TestCase, remove_test_file


class TestAlignedDynamicTableContainer(TestCase):
    """
    Test the AlignedDynamicTable Container class.

    NOTE: Functions specific to linked tables, specifically the:
    * has_foreign_columns
    * get_foreign_columns
    * get_linked_tables
    methods are tested in the test_linkedtables.TestLinkedAlignedDynamicTables class instead of here.
    """
    def setUp(self):
        warnings.simplefilter("always")  # Trigger all warnings
        self.path = 'test_icephys_meta_intracellularrecording.h5'

    def tearDown(self):
        remove_test_file(self.path)

    def test_init(self):
        """Test that just checks that populating the tables with data works correctly"""
        AlignedDynamicTable(
            name='test_aligned_table',
            description='Test aligned container')

    def test_init_categories_without_category_tables_error(self):
        # Test raise error if categories is given without category_tables
        with self.assertRaisesWith(ValueError, "Categories provided but no category_tables given"):
            AlignedDynamicTable(
                name='test_aligned_table',
                description='Test aligned container',
                categories=['cat1', 'cat2'])

    def test_init_length_mismatch_between_categories_and_category_tables(self):
        # Test length mismatch between categories and category_tables
        with self.assertRaisesWith(ValueError,  "0 category_tables given but 2 categories specified"):
            AlignedDynamicTable(
                name='test_aligned_table',
                description='Test aligned container',
                categories=['cat1', 'cat2'],
                category_tables=[])

    def test_init_category_table_names_do_not_match_categories(self):
        # Construct some categories for testing
        category_names = ['test1', 'test2', 'test3']
        num_rows = 10
        categories = [DynamicTable(name=val,
                                   description=val+" description",
                                   columns=[VectorData(name=val+t,
                                                       description=val+t+' description',
                                                       data=np.arange(num_rows)) for t in ['c1', 'c2', 'c3']]
                                   ) for val in category_names]
        # Test add category_table that is not listed in the categories list
        with self.assertRaisesWith(ValueError,
                                   "DynamicTable test3 does not appear in categories ['test1', 'test2', 't3']"):
            AlignedDynamicTable(
                name='test_aligned_table',
                description='Test aligned container',
                categories=['test1', 'test2', 't3'],  # bad name for 'test3'
                category_tables=categories)

    def test_init_duplicate_category_table_name(self):
        # Test duplicate table name
        with self.assertRaisesWith(ValueError, "Duplicate table name test1 found in input dynamic_tables"):
            categories = [DynamicTable(name=val,
                                       description=val+" description",
                                       columns=[VectorData(name=val+t,
                                                           description=val+t+' description',
                                                           data=np.arange(10)) for t in ['c1', 'c2', 'c3']]
                                       ) for val in ['test1', 'test1', 'test3']]
            AlignedDynamicTable(
                name='test_aligned_table',
                description='Test aligned container',
                categories=['test1', 'test2', 'test3'],
                category_tables=categories)

    def test_init_misaligned_category_tables(self):
        """Test misaligned category tables"""
        categories = [DynamicTable(name=val,
                                   description=val+" description",
                                   columns=[VectorData(name=val+t,
                                                       description=val+t+' description',
                                                       data=np.arange(10)) for t in ['c1', 'c2', 'c3']]
                                   ) for val in ['test1', 'test2']]
        categories.append(DynamicTable(name='test3',
                                       description="test3 description",
                                       columns=[VectorData(name='test3 '+t,
                                                           description='test3 '+t+' description',
                                                           data=np.arange(8)) for t in ['c1', 'c2', 'c3']]))
        with self.assertRaisesWith(ValueError,
                                   "Category DynamicTable test3 does not align, it has 8 rows expected 10"):
            AlignedDynamicTable(
                name='test_aligned_table',
                description='Test aligned container',
                categories=['test1', 'test2', 'test3'],
                category_tables=categories)

    def test_init_with_custom_empty_categories(self):
        """Test that we can create an empty table with custom categories"""
        category_names = ['test1', 'test2', 'test3']
        categories = [DynamicTable(name=val, description=val+" description") for val in category_names]
        AlignedDynamicTable(
            name='test_aligned_table',
            description='Test aligned container',
            category_tables=categories)

    def test_init_with_custom_nonempty_categories(self):
        """Test that we can create an empty table with custom categories"""
        category_names = ['test1', 'test2', 'test3']
        num_rows = 10
        categories = [DynamicTable(name=val,
                                   description=val+" description",
                                   columns=[VectorData(name=val+t,
                                                       description=val+t+' description',
                                                       data=np.arange(num_rows)) for t in ['c1', 'c2', 'c3']]
                                   ) for val in category_names]
        temp = AlignedDynamicTable(
            name='test_aligned_table',
            description='Test aligned container',
            category_tables=categories)
        self.assertEqual(temp.categories, category_names)

    def test_init_with_custom_nonempty_categories_and_main(self):
        """
        Test that we can create a non-empty table with custom non-empty categories
        """
        category_names = ['test1', 'test2', 'test3']
        num_rows = 10
        categories = [DynamicTable(name=val,
                                   description=val+" description",
                                   columns=[VectorData(name=t,
                                                       description=val+t+' description',
                                                       data=np.arange(num_rows)) for t in ['c1', 'c2', 'c3']]
                                   ) for val in category_names]
        temp = AlignedDynamicTable(
            name='test_aligned_table',
            description='Test aligned container',
            category_tables=categories,
            columns=[VectorData(name='main_' + t,
                                description='main_'+t+'_description',
                                data=np.arange(num_rows)) for t in ['c1', 'c2', 'c3']])

        self.assertEqual(temp.categories, category_names)
        self.assertTrue('test1' in temp)  # test that contains category works
        self.assertTrue(('test1', 'c1') in temp)  # test that contains a column works
        # test the error case of a tuple with len !=2
        with self.assertRaisesWith(ValueError, "Expected tuple of strings of length 2 got tuple of length 3"):
            ('test1', 'c1', 't3') in temp
        self.assertTupleEqual(temp.colnames, ('main_c1', 'main_c2', 'main_c3'))  # confirm column names

    def test_init_with_custom_misaligned_categories(self):
        """Test that we cannot create an empty table with custom categories"""
        num_rows = 10
        val1 = 'test1'
        val2 = 'test2'
        categories = [DynamicTable(name=val1,
                                   description=val1+" description",
                                   columns=[VectorData(name=val1+t,
                                                       description=val1+t+' description',
                                                       data=np.arange(num_rows)) for t in ['c1', 'c2', 'c3']]),
                      DynamicTable(name=val2,
                                   description=val2+" description",
                                   columns=[VectorData(name=val2+t,
                                                       description=val2+t+' description',
                                                       data=np.arange(num_rows+1)) for t in ['c1', 'c2', 'c3']])
                      ]
        with self.assertRaisesWith(ValueError,
                                   "Category DynamicTable test2 does not align, it has 11 rows expected 10"):
            AlignedDynamicTable(
                name='test_aligned_table',
                description='Test aligned container',
                category_tables=categories)

    def test_init_with_duplicate_custom_categories(self):
        """Test that we can create an empty table with custom categories"""
        category_names = ['test1', 'test1']
        num_rows = 10
        categories = [DynamicTable(name=val,
                                   description=val+" description",
                                   columns=[VectorData(name=val+t,
                                                       description=val+t+' description',
                                                       data=np.arange(num_rows)) for t in ['c1', 'c2', 'c3']]
                                   ) for val in category_names]
        with self.assertRaisesWith(ValueError, "Duplicate table name test1 found in input dynamic_tables"):
            AlignedDynamicTable(
                name='test_aligned_table',
                description='Test aligned container',
                category_tables=categories)

    def test_init_with_bad_custom_categories(self):
        """Test that we cannot provide a category that is not a DynamicTable"""
        num_rows = 10
        categories = [  # good category
                      DynamicTable(name='test1',
                                   description="test1 description",
                                   columns=[VectorData(name='test1'+t,
                                                       description='test1' + t + ' description',
                                                       data=np.arange(num_rows)) for t in ['c1', 'c2', 'c3']]
                                   ),
                      # use a list as a bad category example
                      [0, 1, 2]]
        with self.assertRaisesWith(ValueError, "Category table with index 1 is not a DynamicTable"):
            AlignedDynamicTable(
                name='test_aligned_table',
                description='Test aligned container',
                category_tables=categories)

    def test_round_trip_container(self):
        """Test read and write the container by itself"""
        category_names = ['test1', 'test2', 'test3']
        num_rows = 10
        categories = [DynamicTable(name=val,
                                   description=val+" description",
                                   columns=[VectorData(name=t,
                                                       description=val+t+' description',
                                                       data=np.arange(num_rows)) for t in ['c1', 'c2', 'c3']]
                                   ) for val in category_names]
        curr = AlignedDynamicTable(
            name='test_aligned_table',
            description='Test aligned container',
            category_tables=categories)

        with HDF5IO(self.path, manager=get_manager(), mode='w') as io:
            io.write(curr)

        with HDF5IO(self.path, manager=get_manager(), mode='r') as io:
            incon = io.read()
            self.assertListEqual(incon.categories, curr.categories)
            for n in category_names:
                assert_frame_equal(incon[n], curr[n])

    def test_add_category(self):
        """Test that we can correct a non-empty category to an existing table"""
        category_names = ['test1', 'test2', 'test3']
        num_rows = 10
        categories = [DynamicTable(name=val,
                                   description=val+" description",
                                   columns=[VectorData(name=val+t,
                                                       description=val+t+' description',
                                                       data=np.arange(num_rows)) for t in ['c1', 'c2', 'c3']]
                                   ) for val in category_names]
        adt = AlignedDynamicTable(
            name='test_aligned_table',
            description='Test aligned container',
            category_tables=categories[0:2])
        self.assertListEqual(adt.categories, category_names[0:2])
        adt.add_category(categories[-1])
        self.assertListEqual(adt.categories, category_names)

    def test_add_category_misaligned_rows(self):
        """Test that we can correct a non-empty category to an existing table"""
        category_names = ['test1', 'test2']
        num_rows = 10
        categories = [DynamicTable(name=val,
                                   description=val+" description",
                                   columns=[VectorData(name=val+t,
                                                       description=val+t+' description',
                                                       data=np.arange(num_rows)) for t in ['c1', 'c2', 'c3']]
                                   ) for val in category_names]
        adt = AlignedDynamicTable(
            name='test_aligned_table',
            description='Test aligned container',
            category_tables=categories)
        self.assertListEqual(adt.categories, category_names)
        with self.assertRaisesWith(ValueError, "New category DynamicTable does not align, it has 8 rows expected 10"):
            adt.add_category(DynamicTable(name='test3',
                                          description='test3_description',
                                          columns=[VectorData(name='test3_'+t,
                                                              description='test3 '+t+' description',
                                                              data=np.arange(num_rows - 2)) for t in ['c1', 'c2', 'c3']
                                                   ]))

    def test_add_category_already_in_table(self):
        category_names = ['test1', 'test2', 'test2']
        num_rows = 10
        categories = [DynamicTable(name=val,
                                   description=val+" description",
                                   columns=[VectorData(name=val+t,
                                                       description=val+t+' description',
                                                       data=np.arange(num_rows)) for t in ['c1', 'c2', 'c3']]
                                   ) for val in category_names]
        adt = AlignedDynamicTable(
            name='test_aligned_table',
            description='Test aligned container',
            category_tables=categories[0:2])
        self.assertListEqual(adt.categories, category_names[0:2])
        with self.assertRaisesWith(ValueError, "Category test2 already in the table"):
            adt.add_category(categories[-1])

    def test_add_column(self):
        adt = AlignedDynamicTable(
            name='test_aligned_table',
            description='Test aligned container',
            columns=[VectorData(name='test_'+t,
                                description='test_'+t+' description',
                                data=np.arange(10)) for t in ['c1', 'c2', 'c3']])
        # Test successful add
        adt.add_column(name='testA', description='testA', data=np.arange(10))
        self.assertTupleEqual(adt.colnames,  ('test_c1', 'test_c2', 'test_c3', 'testA'))

    def test_add_column_bad_category(self):
        """Test add column with bad category"""
        adt = AlignedDynamicTable(
            name='test_aligned_table',
            description='Test aligned container',
            columns=[VectorData(name='test_'+t,
                                description='test_'+t+' description',
                                data=np.arange(10)) for t in ['c1', 'c2', 'c3']])
        with self.assertRaisesWith(KeyError, "'Category mycat not in table'"):
            adt.add_column(category='mycat', name='testA', description='testA', data=np.arange(10))

    def test_add_column_bad_length(self):
        """Test add column that is too short"""
        adt = AlignedDynamicTable(
            name='test_aligned_table',
            description='Test aligned container',
            columns=[VectorData(name='test_'+t,
                                description='test_'+t+' description',
                                data=np.arange(10)) for t in ['c1', 'c2', 'c3']])
        # Test successful add
        with self.assertRaisesWith(ValueError, "column must have the same number of rows as 'id'"):
            adt.add_column(name='testA', description='testA', data=np.arange(8))

    def test_add_column_to_subcategory(self):
        """Test adding a column to a subcategory"""
        category_names = ['test1', 'test2', 'test3']
        num_rows = 10
        categories = [DynamicTable(name=val,
                                   description=val+" description",
                                   columns=[VectorData(name=val+t,
                                                       description=val+t+' description',
                                                       data=np.arange(num_rows)) for t in ['c1', 'c2', 'c3']]
                                   ) for val in category_names]
        adt = AlignedDynamicTable(
            name='test_aligned_table',
            description='Test aligned container',
            category_tables=categories)
        self.assertListEqual(adt.categories, category_names)
        # Test successful add
        adt.add_column(category='test2', name='testA', description='testA', data=np.arange(10))
        self.assertTupleEqual(adt.get_category('test2').colnames, ('test2c1', 'test2c2', 'test2c3', 'testA'))

    def test_add_row(self):
        """Test adding a row to a non_empty table"""
        category_names = ['test1', ]
        num_rows = 10
        categories = [DynamicTable(name=val,
                                   description=val+" description",
                                   columns=[VectorData(name=t,
                                                       description=val+t+' description',
                                                       data=np.arange(num_rows)) for t in ['c1', 'c2']]
                                   ) for val in category_names]
        temp = AlignedDynamicTable(
            name='test_aligned_table',
            description='Test aligned container',
            category_tables=categories,
            columns=[VectorData(name='main_' + t,
                                description='main_'+t+'_description',
                                data=np.arange(num_rows)) for t in ['c1', 'c2']])
        self.assertListEqual(temp.categories, category_names)
        # Test successful add
        temp.add_row(test1=dict(c1=1, c2=2), main_c1=3, main_c2=5)
        self.assertListEqual(temp[10].iloc[0].tolist(), [3, 5, 10, 1, 2])
        # Test successful add version 2
        temp.add_row(data=dict(test1=dict(c1=1, c2=2), main_c1=4, main_c2=5))
        self.assertListEqual(temp[11].iloc[0].tolist(), [4, 5, 11, 1, 2])
        # Test missing categories data
        with self.assertRaises(KeyError) as ke:
            temp.add_row(main_c1=3, main_c2=5)
        self.assertTrue("row data keys do not match" in str(ke.exception))

    def test_get_item(self):
        """Test getting elements from the table"""
        category_names = ['test1', ]
        num_rows = 10
        categories = [DynamicTable(name=val,
                                   description=val+" description",
                                   columns=[VectorData(name=t,
                                                       description=val+t+' description',
                                                       data=np.arange(num_rows) + i + 3)
                                            for i, t in enumerate(['c1', 'c2'])]
                                   ) for val in category_names]
        temp = AlignedDynamicTable(
            name='test_aligned_table',
            description='Test aligned container',
            category_tables=categories,
            columns=[VectorData(name='main_' + t,
                                description='main_'+t+'_description',
                                data=np.arange(num_rows)+2) for t in ['c1', 'c2']])
        self.assertListEqual(temp.categories, category_names)
        # Test slicing with a single index
        self.assertListEqual(temp[5].iloc[0].tolist(), [7, 7, 5, 8, 9])
        # Test slice with list
        self.assertListEqual(temp[[5, 7]].iloc[0].tolist(), [7, 7, 5, 8, 9])
        self.assertListEqual(temp[[5, 7]].iloc[1].tolist(), [9, 9, 7, 10, 11])
        # Test slice with slice
        self.assertListEqual(temp[5:7].iloc[0].tolist(), [7, 7, 5, 8, 9])
        self.assertListEqual(temp[5:7].iloc[1].tolist(), [8, 8, 6, 9, 10])
        # Test slice with numpy index arrya
        self.assertListEqual(temp[np.asarray([5, 8])].iloc[0].tolist(), [7, 7, 5, 8, 9])
        self.assertListEqual(temp[np.asarray([5, 8])].iloc[1].tolist(), [10, 10, 8, 11, 12])
        # Test slicing for a single column
        self.assertListEqual(temp['main_c1'][:].tolist(), (np.arange(num_rows)+2).tolist())
        # Test slicing for a single category
        assert_frame_equal(temp['test1'], categories[0].to_dataframe())
        # Test getting the main table
        assert_frame_equal(temp[None], temp.to_dataframe())
        # Test getting a specific column
        self.assertListEqual(temp['test1', 'c1'][:].tolist(), (np.arange(num_rows) + 3).tolist())
        # Test getting a specific cell
        self.assertEqual(temp[None, 'main_c1', 1], 3)
        self.assertEqual(temp[1, None, 'main_c1'], 3)
        # Test bad selection tuple
        with self.assertRaisesWith(ValueError,
                                   "Expected tuple of length 2 of the form [category, column], [row, category], "
                                   "[row, (category, column)] or a tuple of length 3 of the form "
                                   "[category, column, row], [row, category, column]"):
            temp[('main_c1',)]
        # Test selecting a single cell or row of a category table by having a
        # [int, str] or [int, (str, str)] type selection
        # Select row 0 from category 'test1'
        re = temp[0, 'test1']
        self.assertListEqual(re.columns.to_list(), ['id', 'c1', 'c2'])
        self.assertListEqual(re.index.names, [('test_aligned_table', 'id')])
        self.assertListEqual(re.values.tolist()[0], [0, 3, 4])
        # Select a single cell from a columm
        self.assertEqual(temp[1, ('test_aligned_table', 'main_c1')], 3)

    def test_to_dataframe(self):
        """Test that the to_dataframe method works"""
        category_names = ['test1', 'test2', 'test3']
        num_rows = 10
        categories = [DynamicTable(name=val,
                                   description=val+" description",
                                   columns=[VectorData(name=t,
                                                       description=val+t+' description',
                                                       data=np.arange(num_rows)) for t in ['c1', 'c2', 'c3']]
                                   ) for val in category_names]
        adt = AlignedDynamicTable(
            name='test_aligned_table',
            description='Test aligned container',
            category_tables=categories,
            columns=[VectorData(name='main_' + t,
                                description='main_'+t+'_description',
                                data=np.arange(num_rows)) for t in ['c1', 'c2', 'c3']])

        # Test the to_dataframe method with default settings
        tdf = adt.to_dataframe()
        self.assertListEqual(tdf.index.tolist(), list(range(10)))
        self.assertTupleEqual(tdf.index.name, ('test_aligned_table', 'id'))
        expected_cols = [('test_aligned_table', 'main_c1'),
                         ('test_aligned_table', 'main_c2'),
                         ('test_aligned_table', 'main_c3'),
                         ('test1', 'id'), ('test1', 'c1'), ('test1', 'c2'), ('test1', 'c3'),
                         ('test2', 'id'), ('test2', 'c1'), ('test2', 'c2'), ('test2', 'c3'),
                         ('test3', 'id'), ('test3', 'c1'), ('test3', 'c2'), ('test3', 'c3')]
        tdf_cols = tdf.columns.tolist()
        for v in zip(expected_cols, tdf_cols):
            self.assertTupleEqual(v[0], v[1])

        # test the to_dataframe method with ignore_category_ids set to True
        tdf = adt.to_dataframe(ignore_category_ids=True)
        self.assertListEqual(tdf.index.tolist(), list(range(10)))
        self.assertTupleEqual(tdf.index.name, ('test_aligned_table', 'id'))
        expected_cols = [('test_aligned_table', 'main_c1'),
                         ('test_aligned_table', 'main_c2'),
                         ('test_aligned_table', 'main_c3'),
                         ('test1', 'c1'), ('test1', 'c2'), ('test1', 'c3'),
                         ('test2', 'c1'), ('test2', 'c2'), ('test2', 'c3'),
                         ('test3', 'c1'), ('test3', 'c2'), ('test3', 'c3')]
        tdf_cols = tdf.columns.tolist()
        for v in zip(expected_cols, tdf_cols):
            self.assertTupleEqual(v[0], v[1])

    def test_nested_aligned_dynamic_table_not_allowed(self):
        """
        Test that using and AlignedDynamicTable as category for an AlignedDynamicTable is not allowed
        """
        # create an AlignedDynamicTable as category
        subsubcol1 = VectorData(name='sub_sub_column1', description='test sub sub column', data=['test11', 'test12'])
        sub_category = DynamicTable(name='sub_category1', description='test subcategory table', columns=[subsubcol1, ])
        subcol1 = VectorData(name='sub_column1', description='test-subcolumn', data=['test1', 'test2'])
        adt_category = AlignedDynamicTable(
            name='category1',
            description='test using AlignedDynamicTable as a category',
            columns=[subcol1, ],
            category_tables=[sub_category, ])

        # Create a regular column for our main AlignedDynamicTable
        col1 = VectorData(name='column1', description='regular test column', data=['test1', 'test2'])

        # test 1: Make sure we can't add the AlignedDynamicTable category on init
        msg = ("Category table with index %i is an AlignedDynamicTable. "
               "Nesting of AlignedDynamicTable is currently not supported." % 0)
        with self.assertRaisesWith(ValueError, msg):
            # create the nested AlignedDynamicTable with our adt_category as a sub-category
            AlignedDynamicTable(
                name='nested_adt',
                description='test nesting AlignedDynamicTable',
                columns=[col1, ],
                category_tables=[adt_category, ])

        # test 2: Make sure we can't add the AlignedDynamicTable category via add_category
        adt = AlignedDynamicTable(
            name='nested_adt',
            description='test nesting AlignedDynamicTable',
            columns=[col1, ])
        msg = "Category is an AlignedDynamicTable. Nesting of AlignedDynamicTable is currently not supported."
        with self.assertRaisesWith(ValueError, msg):
            adt.add_category(adt_category)

    def test_dynamictable_region_to_aligneddynamictable(self):
        """
        Test to ensure data is being retrieved correctly when pointing to an AlignedDynamicTable.
        In particular, make sure that all columns are being used, including those of the
        category tables, not just the ones from the main table.
        """
        temp_table = DynamicTable(name='t1', description='t1',
                                  colnames=['c1', 'c2'],
                                  columns=[VectorData(name='c1', description='c1', data=np.arange(4)),
                                           VectorData(name='c2', description='c2', data=np.arange(4))])
        temp_aligned_table = AlignedDynamicTable(name='my_aligned_table',
                                                 description='my test table',
                                                 category_tables=[temp_table],
                                                 colnames=['a1', 'a2'],
                                                 columns=[VectorData(name='a1', description='c1', data=np.arange(4)),
                                                          VectorData(name='a2', description='c1', data=np.arange(4))])
        dtr = DynamicTableRegion(name='test', description='test', data=np.arange(4), table=temp_aligned_table)
        dtr_df = dtr[:]
        # Full number of rows
        self.assertEqual(len(dtr_df), 4)
        # Test num columns: 2 columns from the main table, 2 columns from the category, 1 id columns from the category
        self.assertEqual(len(dtr_df.columns), 5)
        # Test that the data is correct
        for i, v in enumerate([('my_aligned_table', 'a1'), ('my_aligned_table', 'a2'),
                               ('t1', 'id'), ('t1', 'c1'), ('t1', 'c2')]):
            self.assertTupleEqual(dtr_df.columns[i], v)
        # Test the column data
        for c in dtr_df.columns:
            self.assertListEqual(dtr_df[c].to_list(), list(range(4)))

    def test_get_colnames(self):
        """
        Test the AlignedDynamicTable.get_colnames function
        """
        category_names = ['test1', 'test2', 'test3']
        num_rows = 10
        categories = [DynamicTable(name=val,
                                   description=val+" description",
                                   columns=[VectorData(name=t,
                                                       description=val+t+' description',
                                                       data=np.arange(num_rows)) for t in ['c1', 'c2', 'c3']]
                                   ) for val in category_names]
        adt = AlignedDynamicTable(
            name='test_aligned_table',
            description='Test aligned container',
            category_tables=categories,
            columns=[VectorData(name='main_' + t,
                                description='main_'+t+'_description',
                                data=np.arange(num_rows)) for t in ['c1', 'c2', 'c3']])
        # Default, only get the colnames of the main table. Same as adt.colnames property
        expected_colnames = ('main_c1', 'main_c2', 'main_c3')
        self.assertTupleEqual(adt.get_colnames(), expected_colnames)
        # Same as default because if we don't include the catgories than ignore_category_ids has no effect
        self.assertTupleEqual(adt.get_colnames(include_category_tables=False, ignore_category_ids=True),
                              expected_colnames)
        # Full set of columns
        expected_colnames = [('test_aligned_table', 'main_c1'), ('test_aligned_table', 'main_c2'),
                             ('test_aligned_table', 'main_c3'), ('test1', 'id'), ('test1', 'c1'),
                             ('test1', 'c2'), ('test1', 'c3'), ('test2', 'id'), ('test2', 'c1'),
                             ('test2', 'c2'), ('test2', 'c3'), ('test3', 'id'), ('test3', 'c1'),
                             ('test3', 'c2'), ('test3', 'c3')]
        self.assertListEqual(adt.get_colnames(include_category_tables=True, ignore_category_ids=False),
                             expected_colnames)
        # All columns without the id columns of the category tables
        expected_colnames = [('test_aligned_table', 'main_c1'), ('test_aligned_table', 'main_c2'),
                             ('test_aligned_table', 'main_c3'), ('test1', 'c1'), ('test1', 'c2'),
                             ('test1', 'c3'), ('test2', 'c1'), ('test2', 'c2'), ('test2', 'c3'),
                             ('test3', 'c1'), ('test3', 'c2'), ('test3', 'c3')]
        self.assertListEqual(adt.get_colnames(include_category_tables=True, ignore_category_ids=True),
                             expected_colnames)



import json
import os
from argparse import ArgumentParser
from glob import glob

import jsonschema
import ruamel.yaml as yaml


def validate_spec(fpath_spec, fpath_schema):
    """
    Validate a yaml specification file against the json schema file that
    defines the specification language. Can be used to validate changes
    to the NWB and HDMF core schemas, as well as any extensions to either.

    :param fpath_spec: path-like
    :param fpath_schema: path-like
    """

    schemaAbs = 'file://' + os.path.abspath(fpath_schema)

    f_schema = open(fpath_schema, 'r')
    schema = json.load(f_schema)

    class FixResolver(jsonschema.RefResolver):
        def __init__(self):
            jsonschema.RefResolver.__init__(self,
                                            base_uri=schemaAbs,
                                            referrer=None)
            self.store[schemaAbs] = schema

    new_resolver = FixResolver()

    f_nwb = open(fpath_spec, 'r')
    instance = yaml.safe_load(f_nwb)

    jsonschema.validate(instance, schema, resolver=new_resolver)


def main():
    parser = ArgumentParser(description="Validate an HDMF/NWB specification")
    parser.add_argument("paths", type=str, nargs='+', help="yaml file paths")
    parser.add_argument("-m", "--metaschema", type=str,
                        help=".json.schema file used to validate yaml files")
    args = parser.parse_args()

    for path in args.paths:
        if os.path.isfile(path):
            validate_spec(path, args.metaschema)
        elif os.path.isdir(path):
            for ipath in glob(os.path.join(path, '*.yaml')):
                validate_spec(ipath, args.metaschema)
        else:
            raise ValueError('path must be a valid file or directory')


if __name__ == "__main__":
    main()



import numpy as np
import os
import re
import unittest
from abc import ABCMeta, abstractmethod

from .utils import remove_test_file
from ..backends.hdf5 import HDF5IO
from ..build import Builder
from ..common import validate as common_validate, get_manager
from ..container import AbstractContainer, Container, Data
from ..utils import get_docval_macro
from ..data_utils import AbstractDataChunkIterator


class TestCase(unittest.TestCase):
    """
    Extension of unittest's TestCase to add useful functions for unit testing in HDMF.
    """

    def assertRaisesWith(self, exc_type, exc_msg, *args, **kwargs):
        """
        Asserts the given invocation raises the expected exception. This is similar to unittest's assertRaises and
        assertRaisesRegex, but checks for an exact match.
        """

        return self.assertRaisesRegex(exc_type, '^%s$' % re.escape(exc_msg), *args, **kwargs)

    def assertWarnsWith(self, warn_type, exc_msg, *args, **kwargs):
        """
        Asserts the given invocation raises the expected warning. This is similar to unittest's assertWarns and
        assertWarnsRegex, but checks for an exact match.
        """

        return self.assertWarnsRegex(warn_type, '^%s$' % re.escape(exc_msg), *args, **kwargs)

    def assertContainerEqual(self,
                             container1,
                             container2,
                             ignore_name=False,
                             ignore_hdmf_attrs=False,
                             ignore_string_to_byte=False,
                             message=None):
        """
        Asserts that the two AbstractContainers have equal contents. This applies to both Container and Data types.

        :param container1: First container
        :type container1: AbstractContainer
        :param container2: Second container to compare with container 1
        :type container2: AbstractContainer
        :param ignore_name: whether to ignore testing equality of name of the top-level container
        :param ignore_hdmf_attrs: whether to ignore testing equality of HDMF container attributes, such as
                                  container_source and object_id
        :param ignore_string_to_byte: ignore conversion of str to bytes and compare as unicode instead
        :param message: custom additional message to show when assertions as part of this assert are failing
        """
        self.assertTrue(isinstance(container1, AbstractContainer), message)
        self.assertTrue(isinstance(container2, AbstractContainer), message)
        type1 = type(container1)
        type2 = type(container2)
        self.assertEqual(type1, type2, message)
        if not ignore_name:
            self.assertEqual(container1.name, container2.name, message)
        if not ignore_hdmf_attrs:
            self.assertEqual(container1.container_source, container2.container_source, message)
            self.assertEqual(container1.object_id, container2.object_id, message)
        # NOTE: parent is not tested because it can lead to infinite loops
        if isinstance(container1, Container):
            self.assertEqual(len(container1.children), len(container2.children), message)
        # do not actually check the children values here. all children *should* also be fields, which is checked below.
        # this is in case non-field children are added to one and not the other

        for field in getattr(container1, type1._fieldsname):
            with self.subTest(field=field, container_type=type1.__name__):
                f1 = getattr(container1, field)
                f2 = getattr(container2, field)
                self._assert_field_equal(f1, f2,
                                         ignore_hdmf_attrs=ignore_hdmf_attrs,
                                         ignore_string_to_byte=ignore_string_to_byte,
                                         message=message)

    def _assert_field_equal(self,
                            f1,
                            f2,
                            ignore_hdmf_attrs=False,
                            ignore_string_to_byte=False,
                            message=None):
        """
        Internal helper function used to compare two fields from Container objects

        :param f1: The first field
        :param f2: The second field
        :param ignore_hdmf_attrs: whether to ignore testing equality of HDMF container attributes, such as
                                  container_source and object_id
        :param ignore_string_to_byte: ignore conversion of str to bytes and compare as unicode instead
        :param message: custom additional message to show when assertions as part of this assert are failing
        """
        array_data_types = get_docval_macro('array_data')
        if (isinstance(f1, array_data_types) or isinstance(f2, array_data_types)):
            self._assert_array_equal(f1, f2,
                                     ignore_hdmf_attrs=ignore_hdmf_attrs,
                                     ignore_string_to_byte=ignore_string_to_byte,
                                     message=message)
        elif isinstance(f1, dict) and len(f1) and isinstance(f1.values()[0], Container):
            self.assertIsInstance(f2, dict, message)
            f1_keys = set(f1.keys())
            f2_keys = set(f2.keys())
            self.assertSetEqual(f1_keys, f2_keys, message)
            for k in f1_keys:
                with self.subTest(module_name=k):
                    self.assertContainerEqual(f1[k], f2[k],
                                              ignore_hdmf_attrs=ignore_hdmf_attrs,
                                              ignore_string_to_byte=ignore_string_to_byte,
                                              message=message)
        elif isinstance(f1, Container):
            self.assertContainerEqual(f1, f2,
                                      ignore_hdmf_attrs=ignore_hdmf_attrs,
                                      ignore_string_to_byte=ignore_string_to_byte,
                                      message=message)
        elif isinstance(f1, Data):
            self._assert_data_equal(f1, f2,
                                    ignore_hdmf_attrs=ignore_hdmf_attrs,
                                    ignore_string_to_byte=ignore_string_to_byte,
                                    message=message)
        elif isinstance(f1, (float, np.floating)):
            np.testing.assert_allclose(f1, f2, err_msg=message)
        else:
            self.assertEqual(f1, f2, message)

    def _assert_data_equal(self,
                           data1,
                           data2,
                           ignore_hdmf_attrs=False,
                           ignore_string_to_byte=False,
                           message=None):
        """
        Internal helper function used to compare two :py:class:`~hdmf.container.Data` objects

        :param data1: The first :py:class:`~hdmf.container.Data` object
        :type data1: :py:class:`hdmf.container.Data`
        :param data1: The second :py:class:`~hdmf.container.Data` object
        :type data1: :py:class:`hdmf.container.Data
        :param ignore_hdmf_attrs: whether to ignore testing equality of HDMF container attributes, such as
                                  container_source and object_id
        :param ignore_string_to_byte: ignore conversion of str to bytes and compare as unicode instead
        :param message: custom additional message to show when assertions as part of this assert are failing
        """
        self.assertTrue(isinstance(data1, Data), message)
        self.assertTrue(isinstance(data2, Data), message)
        self.assertEqual(len(data1), len(data2), message)
        self._assert_array_equal(data1.data, data2.data,
                                 ignore_hdmf_attrs=ignore_hdmf_attrs,
                                 ignore_string_to_byte=ignore_string_to_byte,
                                 message=message)
        self.assertContainerEqual(container1=data1,
                                  container2=data2,
                                  ignore_hdmf_attrs=ignore_hdmf_attrs,
                                  message=message)

    def _assert_array_equal(self,
                            arr1,
                            arr2,
                            ignore_hdmf_attrs=False,
                            ignore_string_to_byte=False,
                            message=None):
        """
        Internal helper function used to check whether two arrays are equal

        :param arr1: The first array
        :param arr2: The second array
        :param ignore_hdmf_attrs: whether to ignore testing equality of HDMF container attributes, such as
                                  container_source and object_id
        :param ignore_string_to_byte: ignore conversion of str to bytes and compare as unicode instead
        :param message: custom additional message to show when assertions as part of this assert are failing
        """
        array_data_types = tuple([i for i in get_docval_macro('array_data')
                                  if (i != list and i != tuple and i != AbstractDataChunkIterator)])
        # We construct array_data_types this way to avoid explicit dependency on h5py, Zarr and other
        # I/O backends. Only list and tuple do not support [()] slicing, and AbstractDataChunkIterator
        # should never occur here. The effective value of array_data_types is then:
        # array_data_types = (np.ndarray, h5py.Dataset, zarr.core.Array, hdmf.query.HDMFDataset)
        if isinstance(arr1, array_data_types):
            arr1 = arr1[()]
        if isinstance(arr2, array_data_types):
            arr2 = arr2[()]
        if not isinstance(arr1, (tuple, list, np.ndarray)) and not isinstance(arr2, (tuple, list, np.ndarray)):
            if isinstance(arr1, (float, np.floating)):
                np.testing.assert_allclose(arr1, arr2, err_msg=message)
            else:
                if ignore_string_to_byte:
                    if isinstance(arr1, bytes):
                        arr1 = arr1.decode('utf-8')
                    if isinstance(arr2, bytes):
                        arr2 = arr2.decode('utf-8')
                self.assertEqual(arr1, arr2, message)  # scalar
        else:
            self.assertEqual(len(arr1), len(arr2), message)
            if isinstance(arr1, np.ndarray) and len(arr1.dtype) > 1:  # compound type
                arr1 = arr1.tolist()
            if isinstance(arr2, np.ndarray) and len(arr2.dtype) > 1:  # compound type
                arr2 = arr2.tolist()
            if isinstance(arr1, np.ndarray) and isinstance(arr2, np.ndarray):
                if np.issubdtype(arr1.dtype, np.number):
                    np.testing.assert_allclose(arr1, arr2, err_msg=message)
                else:
                    np.testing.assert_array_equal(arr1, arr2, err_msg=message)
            else:
                for sub1, sub2 in zip(arr1, arr2):
                    if isinstance(sub1, Container):
                        self.assertContainerEqual(sub1, sub2,
                                                  ignore_hdmf_attrs=ignore_hdmf_attrs,
                                                  ignore_string_to_byte=ignore_string_to_byte,
                                                  message=message)
                    elif isinstance(sub1, Data):
                        self._assert_data_equal(sub1, sub2,
                                                ignore_hdmf_attrs=ignore_hdmf_attrs,
                                                ignore_string_to_byte=ignore_string_to_byte,
                                                message=message)
                    else:
                        self._assert_array_equal(sub1, sub2,
                                                 ignore_hdmf_attrs=ignore_hdmf_attrs,
                                                 ignore_string_to_byte=ignore_string_to_byte,
                                                 message=message)

    def assertBuilderEqual(self,
                           builder1,
                           builder2,
                           check_path=True,
                           check_source=True,
                           message=None):
        """
        Test whether two builders are equal. Like assertDictEqual but also checks type, name, path, and source.

        :param builder1: The first builder
        :type builder1: Builder
        :param builder2: The second builder
        :type builder2: Builder
        :param check_path: Check that the builder.path values are equal
        :type check_path: bool
        :param check_source: Check that the builder.source values are equal
        :type check_source: bool
        :param message: Custom message to add when any asserts as part of this assert are failing
        :type message: str or None (default=None)
        """
        self.assertTrue(isinstance(builder1, Builder), message)
        self.assertTrue(isinstance(builder2, Builder), message)
        self.assertEqual(type(builder1), type(builder2), message)
        self.assertEqual(builder1.name, builder2.name, message)
        if check_path:
            self.assertEqual(builder1.path, builder2.path, message)
        if check_source:
            self.assertEqual(builder1.source, builder2.source, message)
        self.assertDictEqual(builder1, builder2, message)


class H5RoundTripMixin(metaclass=ABCMeta):
    """
    Mixin class for methods to run a roundtrip test writing a container to and reading the container from an HDF5 file.
    The setUp, test_roundtrip, and tearDown methods will be run by unittest.

    The abstract method setUpContainer needs to be implemented by classes that include this mixin.

    Example::

        class TestMyContainerRoundTrip(H5RoundTripMixin, TestCase):
            def setUpContainer(self):
                # return the Container to read/write

    NOTE: This class is a mix-in and not a subclass of TestCase so that unittest does not discover it, try to run it,
    and skip it.
    """

    def setUp(self):
        self.__manager = get_manager()
        self.container = self.setUpContainer()
        self.container_type = self.container.__class__.__name__
        self.filename = 'test_%s.h5' % self.container_type
        self.export_filename = 'test_export_%s.h5' % self.container_type
        self.writer = None
        self.reader = None
        self.export_reader = None

    def tearDown(self):
        if self.writer is not None:
            self.writer.close()
        if self.reader is not None:
            self.reader.close()
        if self.export_reader is not None:
            self.export_reader.close()

        remove_test_file(self.filename)
        remove_test_file(self.export_filename)

    @abstractmethod
    def setUpContainer(self):
        """Return the Container to read/write."""
        raise NotImplementedError('Cannot run test unless setUpContainer is implemented')

    def test_roundtrip(self):
        """Test whether the container read from a written file is the same as the original file."""
        read_container = self.roundtripContainer()
        self._test_roundtrip(read_container, export=False)

    def test_roundtrip_export(self):
        """Test whether the container read from a written and then exported file is the same as the original file."""
        read_container = self.roundtripExportContainer()
        self._test_roundtrip(read_container, export=True)

    def _test_roundtrip(self, read_container, export=False):
        self.assertIsNotNone(str(self.container))  # added as a test to make sure printing works
        self.assertIsNotNone(str(read_container))
        # make sure we get a completely new object
        self.assertNotEqual(id(self.container), id(read_container))
        # the name of the root container of a file is always 'root' (see h5tools.py ROOT_NAME)
        # thus, ignore the name of the container when comparing original container vs read container
        if not export:
            self.assertContainerEqual(read_container, self.container, ignore_name=True)
        else:
            self.assertContainerEqual(read_container, self.container, ignore_name=True, ignore_hdmf_attrs=True)

        self.validate(read_container._experimental)

    def roundtripContainer(self, cache_spec=False):
        """Write the container to an HDF5 file, read the container from the file, and return it."""
        with HDF5IO(self.filename, manager=get_manager(), mode='w') as write_io:
            write_io.write(self.container, cache_spec=cache_spec)

        self.reader = HDF5IO(self.filename, manager=get_manager(), mode='r')
        return self.reader.read()

    def roundtripExportContainer(self, cache_spec=False):
        """Write the container to an HDF5 file, read it, export it to a new file, read that file, and return it."""
        self.roundtripContainer(cache_spec=cache_spec)

        HDF5IO.export_io(
            src_io=self.reader,
            path=self.export_filename,
            cache_spec=cache_spec,
        )

        self.export_reader = HDF5IO(self.export_filename, manager=get_manager(), mode='r')
        return self.export_reader.read()

    def validate(self, experimental=False):
        """Validate the written and exported files, if they exist."""
        if os.path.exists(self.filename):
            with HDF5IO(self.filename, manager=get_manager(), mode='r') as io:
                errors = common_validate(io, experimental=experimental)
                if errors:
                    for err in errors:
                        raise Exception(err)

        if os.path.exists(self.export_filename):
            with HDF5IO(self.filename, manager=get_manager(), mode='r') as io:
                errors = common_validate(io, experimental=experimental)
                if errors:
                    for err in errors:
                        raise Exception(err)



from . import query  # noqa: F401
from .container import Container, Data, DataRegion
from .utils import docval, getargs
from .region import ListSlicer
from .backends.hdf5.h5_utils import H5RegionSlicer, H5Dataset


@docval({'name': 'dataset', 'type': None, 'doc': 'the HDF5 dataset to slice'},
        {'name': 'region', 'type': None, 'doc': 'the region reference to use to slice'},
        is_method=False)
def get_region_slicer(**kwargs):
    import warnings  # noqa: E402
    warnings.warn('get_region_slicer is deprecated and will be removed in HDMF 3.0.', DeprecationWarning)

    dataset, region = getargs('dataset', 'region', kwargs)
    if isinstance(dataset, (list, tuple, Data)):
        return ListSlicer(dataset, region)
    elif isinstance(dataset, H5Dataset):
        return H5RegionSlicer(dataset, region)
    return None


from . import _version  # noqa: F401,E402
__version__ = _version.get_versions()['version']


from ._due import due, BibTeX  # noqa: E402
due.cite(BibTeX("""
@INPROCEEDINGS{9005648,
  author={A. J. {Tritt} and O. {Rübel} and B. {Dichter} and R. {Ly} and D. {Kang} and E. F. {Chang} and L. M. {Frank} and K. {Bouchard}},
  booktitle={2019 IEEE International Conference on Big Data (Big Data)},
  title={HDMF: Hierarchical Data Modeling Framework for Modern Science Data Standards},
  year={2019},
  volume={},
  number={},
  pages={165-179},
  doi={10.1109/BigData47090.2019.9005648}}
"""), description="HDMF: Hierarchical Data Modeling Framework for Modern Science Data Standards",  # noqa: E501
         path="hdmf/", version=__version__, cite_module=True)
del due, BibTeX
